{******************************************************************************}
{                                                                              }
{ Event Tracing for Windows consumer Unit                                      }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License          }
{ Version 1.1 (the "License"); you may not use this file except in             }
{ compliance with the License. You may obtain a copy of the License at         }
{ http://www.mozilla.org/MPL/                                                  }
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS"          }
{ basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the      }
{ License for the specific language governing rights and limitations           }
{ under the License.                                                           }
{                                                                              }
{ The Original Code is uETWAdvClient.pas released February 2008.                  }
{                                                                              }
{ The Initial Developer of the Original Code is Franck SORIANO.                }
{                                                                              }
{ Alternatively, the contents of this file may be used under the terms of the  }
{ GNU Lesser General Public License (the  "LGPL License"), in which case the   }
{ provisions of the LGPL License are applicable instead of those above.        }
{ If you wish to allow use of your version of this file only under the terms   }
{ of the LGPL License and not to allow others to use your version of this file }
{ under the MPL, indicate your decision by deleting  the provisions above and  }
{ replace  them with the notice and other provisions required by the LGPL      }
{ License.  If you do not delete the provisions above, a recipient may use     }
{ your version of this file under either the MPL or the LGPL License.          }
{                                                                              }
{ For more information about the LGPL: http://www.gnu.org/copyleft/lesser.html }
{                                                                              }
{******************************************************************************}
unit uETWAdvClient;

interface

uses classes, sysutils, windows, 
     uETW, uETWClient, uETWWmiInfo;

type
  // Cette classe implmente un client ETW gnrique capable de traiter
  // n'importe quel type de trace, avec des vnements provenants de n'importe
  // quel provider.
  TETWAdvClient = class(TETWClient)
  private
    FThreadCol : TList;             // Liste des threads qui ont gnrs un
                                    // vnement dans la trace.
  protected
    // ProcessEvent est appele par la fonction callback chaque fois qu'un
    // vnement de la trace est lu. Elle traite l'vnement reu en ajoutant
    // une instance de la classe TTraceData dans la trace.
    procedure ProcessEvent(pEvent : PEVENT_TRACE); override;

  public
    // Cre et initialise une nouvelle instance de la classe TETWAdvClient.
    constructor Create(Name : string; IsRealTime : boolean);

    // Permet d'enregistrer la trace dans un fichier txt.
    procedure SaveToFile(fileName : string);

    // Efface la liste des Threads.
    procedure ResetThreadCol;

    // Retourne le rang du thread identifi par son identifiant ThreadId dans la
    // trace. Le premier thread qui crit un vnement possde le rang 0.
    // Le deuxime le rang 1...
    // On peut s'en servir pour formatter la trace, par exemple en affichant une
    // colonne par thread.
    function GetThreadCol(ThreadId : integer) : integer;

    // Retourne le nombre de thread ayant crit des vnements dans la trace
    // depuis le dernier appel  ResetThreadCol.
    function GetNbThread : integer;

    // Efface les vnements de la trace.
    procedure ClearTrace; override;
  end;

  
implementation
uses uTraceData; // uTraceData contient la dclaration de l'objet TTraceData qui
                 // modlise un vnement de la trace.

// Cre et initialise une nouvelle instance de la classe TETWAdvClient.
procedure TETWAdvClient.ClearTrace;
begin
  inherited ClearTrace;
  FThreadCol.Clear;
end;

constructor TETWAdvClient.Create(Name: string; IsRealTime: boolean);
begin
  FThreadCol := TList.Create;

  inherited Create(Name, IsRealTime);
end;

function TETWAdvClient.GetNbThread: integer;
begin
  result := FThreadCol.Count;
end;

// Retourne le rang du thread identifi par son identifiant ThreadId dans la
// trace. Le premier thread qui crit un vnement possde le rang 0.
// Le deuxime le rang 1...
// On peut s'en servir pour formatter la trace, par exemple en affichant une
// colonne par thread.
function TETWAdvClient.GetThreadCol(ThreadId : integer) : integer;
var i : integer;
begin
  result := -1;
  // La valeur retourne est tout simplement l'index de <ThreadId> dans
  // FThreadCol. Donc on recherche <ThreadId> dans la liste, avec un simple
  // scan squentiel. En effet, on s'attend  avoir moins d'une dizaine de thread
  // dans la liste.
  for i := 0 to FThreadCol.Count -1 do
  begin
    if ThreadId = integer(FThreadCol[i])
    then begin
      result := i;
      break;
    end;
  end;

  if result = -1 // On regarde si le thread a t trouv
  then begin
    // Le thread n'existe pas, on l'ajoute  la liste.
    FThreadCol.Add(pointer(ThreadId));
    result := FThreadCol.Count-1;
  end;
end;

// ProcessEvent est appele par la fonction callback chaque fois qu'un
// vnement de la trace est lu. Cette mthode s'occupe de dcoder
// l'enregistrement et de le mmoriser.
procedure TETWAdvClient.ProcessEvent(pEvent: PEVENT_TRACE);
var
  Data : TTraceData;
begin
  Data := TTraceData.Create(pEvent);
  Data.ThreadCol := GetThreadCol(Data.ThreadId);
  Trace.Add(Data);
end;

// Efface la liste des Threads.
procedure TETWAdvClient.ResetThreadCol;
begin
  FThreadCol.Clear;
end;

// Permet d'enregistrer la trace dans un fichier txt.
procedure TETWAdvClient.SaveToFile(fileName: string);
var lst : TStrings;
    i,j : integer;
    s : string;
    Data : TTraceData;
    delta : int64;
    LastTimeStamp : int64;
    wmi : TWmiObject;
    offset : cardinal;
    Field : TMOFField;
    FieldStr : string;
begin
  if Trace.Count > 0 // On n'exporte pas la trace si elle est vide.
  then begin
    lst := TStringList.Create;
    try
      // Ecriture de l'entete. Les colonnes sont spares par des tabulations.
      lst.Add('ClockTime'#9'Date'#9'Delta'#9'Event'#9'Type'#9'PID'#9'TID'#9'User');
      LastTimeStamp := TTraceData(Trace[0]).TimeStamp;
      for i := 0 to Trace.Count -1 do
      begin
        Data := Trace[i];
        s := '';
        // On calcul le temps coul depuis la ligne prcdente.
        Delta := Data.TimeStamp - LastTimeStamp;
        LastTimeStamp := Data.TimeStamp;

        // On formate la partie fixe de l'vnement
        s := QuotedStr(IntToStr(Data.TimeStamp)) + #9 + // ClockTime brut
             // On convertit ClockTime en date/heure
             QuotedStr(DateTimeToStr(GetDateTime(Data.TimeStamp))) + #9 +
             Format('%.4f', [Delta / 10000]) + #9 + // Temps coul en ms.
             QuotedStr(Data.GetEventClass) + #9 + // Nom de la classe d'vnement
             QuotedStr(Data.GetEventType)+ #9 + // Nom du type de l'vnement
             IntToStr(Data.PID)+ #9 + // Identifiant du process d'origine
             IntToStr(Data.ThreadId)+ #9; // Identifiant du thread d'origine.

        // Ajout des donnes utilisateur, selon les informations MOF publies.
        wmi := Data.GetInfo;
        if Assigned(wmi)
        then begin
          // Si on ne trouve pas d'informations MOF l'vnement est ignor
          offset := 0;
          for j := 0 to wmi.Properties.Count-1 do
          begin
            Field := TMofField(wmi.Properties.Objects[j]);

            // On formate le champ dans un string. GetValue doit tre appele
            // mme si le champ n'est pas crit dans la trace. En effet c'est
            // GetValue qui positionne Offset et permet de passer au champ
            // suivant.
            FieldStr := Field.GetValue(@Data.UserData[0],
                                       High(Data.UserData)+1,
                                       Offset);

            // Les proprits marques avec l'extension NoPrint ne sont pas
            // affiches, donc on ne les exporte pas.
            if not SameText(Field.Extension, 'NoPrint')
            then begin
              // Avant d'crire la ligne, on se protge des caractres de
              // tabulation qui seraient interprt comme un saut de colonne.
              FieldStr := StringReplace(FieldStr, #9, ' ', [rfReplaceAll]);

              // On se protge galement des sauts de lignes qui indiqueraient
              // la fin de l'enregistrement.
              FieldStr := StringReplace(FieldStr, #13#10, ' ', [rfReplaceAll]);              
              FieldStr := StringReplace(FieldStr, #13, ' ', [rfReplaceAll]);
              FieldStr := StringReplace(FieldStr, #10, ' ', [rfReplaceAll]);              
              s := s + FieldStr + #9;
            end;
          end;

          // On regarde s'il faut supprimer la tabulation finale.
          if wmi.Properties.Count>0
          then SetLength(s, length(s)-1);
        end;
        lst.add(s);
      end;

      // Ecriture de la trace sur disque.
      lst.SaveToFile(FileName);
    finally
      lst.free;
    end;
  end;
end;

end.
