{******************************************************************************}
{                                                                              }
{ Event Tracing for Windows consumer Unit                                      }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License          }
{ Version 1.1 (the "License"); you may not use this file except in             }
{ compliance with the License. You may obtain a copy of the License at         }
{ http://www.mozilla.org/MPL/                                                  }
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS"          }
{ basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the      }
{ License for the specific language governing rights and limitations           }
{ under the License.                                                           }
{                                                                              }
{ The Original Code is uETWClient.pas released February 2008.                  }
{                                                                              }
{ The Initial Developer of the Original Code is Franck SORIANO.                }
{                                                                              }
{ Alternatively, the contents of this file may be used under the terms of the  }
{ GNU Lesser General Public License (the  "LGPL License"), in which case the   }
{ provisions of the LGPL License are applicable instead of those above.        }
{ If you wish to allow use of your version of this file only under the terms   }
{ of the LGPL License and not to allow others to use your version of this file }
{ under the MPL, indicate your decision by deleting  the provisions above and  }
{ replace  them with the notice and other provisions required by the LGPL      }
{ License.  If you do not delete the provisions above, a recipient may use     }
{ your version of this file under either the MPL or the LGPL License.          }
{                                                                              }
{ For more information about the LGPL: http://www.gnu.org/copyleft/lesser.html }
{                                                                              }
{******************************************************************************}
unit uETWClient;

interface
uses classes, sysutils, windows,
     uETW, uEventTrace;

type
  // TETWClient
  // Cette classe permet de lire une trace existante et traiter ses vnements.
  TETWClient = class(TObject)
  private
    FWorkerThread : TThread;
    FTrace : EVENT_TRACE_LOGFILE;   // Entte de la trace qui a t ouverte.
    FList : TList;                  // Liste des vnements de la trace.
    FSession : TRACEHANDLE;         // Handle de la trace ouverte.
    FTraceMutex : Cardinal;
  protected
    // ProcessEvent est appele par la fonction callback chaque fois qu'un
    // vnement de la trace est lu. Cette mthode s'occupe de dcoder
    // l'enregistrement et de le mmoriser.
    procedure ProcessEvent(pEvent : PEVENT_TRACE); virtual;

  public
    // Cre et initialise une nouvelle instance de la classe TETWClient :
    // Ouvre une trace existante et commence  lire ses vnement. On peut lire
    // une trace fichier ou une trace temps rel.
    // Si on ouvre une trace fichier, le constructeur rend la main une fois que
    // tous les vnements de la trace ont t lus.
    // Si on ouvre une trace temps rel, le constructeur rend immdiatement la
    // main, les vnements seront lus au fur et  mesure de leur rception en
    // tche de fond.
    // <Name> : Si on veut ouvrir une trace fichier, <Name> doit contenir le nom
    //          du fichier .etl  ouvrir. Si on veut ouvrir une trace temps rel,
    //          <Name> doit contenir le nom de la session de trace  ouvrir.
    // <IsRealTime> : Indique si on veut ouvrir une trace temps rel (true) ou
    //                une trace fichier (false).
    // ATTENTION: TETWClient ne peut traiter qu'une seule trace  la fois.
    //            Ca ne pose pas de problme avec les traces fichiers puisque
    //            le constructeur ne rend la main qu'une fois le traitement de
    //            la trace termin.
    //            En revanche, si on lit une trace temps rel, on ne pourra pas
    //            lire d'autres traces tant que la trace n'aura pas t arrte !  
    constructor Create(Name : string; IsRealTime : boolean);

    // Dtruit la classe et libre les ressources.
    destructor Destroy; override;

    // Fonction utilitaire permettant de convertir les timestamps de la trace
    // en Date et Heure Delphi.
    function GetDateTime(ClockTime : int64) : TDateTime;

    // Efface les vnements de la trace.
    procedure ClearTrace; virtual;

    // Ajoute l'vnement Event dans la trace. Pour les traces temps rels, il
    // est possible qu'on flush les buffers dans un ordre diffrent de
    // l'ordre chronologique. De sorte qu'on peut recevoir les vnements
    // dans le dsordre, mme si ETW les trie avant de nous les envoyer.
    // C'est pourqoi AddEvent tri la trace, en faisant attention  insrer
    // chaque vnement  la bonne position chronologique.  
    procedure AddEvent(Event : TEventTrace);

    // Indique que l'on veut modifier la trace. Cette mthode permet d'obtenir
    // un acces exclusif  la trace.
    procedure Lock;

    // Indique que l'on a finit de modifier la trace.
    procedure UnLock;

    // Trie la trace par ordre chronologique.
    procedure SortTrace;

    procedure StopTrace;

    // Liste des vnements lus depuis la trace.
    property Trace : TList read FList;
  end;

implementation
uses uSQLLogger, uLogger;


// TimeDay : Dure d'une journe pour la conversion des timestamps
const TimeDay = 864000000000;
      T0 = 109205 * TimeDay; // Dure du dcalage entre la rfrence des dates
                             // Delphi et la rfrence des dates ETW.

// ActiveClient est un pointeur sur l'instance TETWClient en cours de traitement
// La variable est utilise temporairement pendant le traitement d'une trace
// fichier. Pour une trace temps rel, ActiveClient est utilis jusqu' l'arrt
// de la trace.
var ActiveClient : TETWClient = nil;

// TWorkerThread est utilis pour traiter la trace temps rel. En effet
// ProcessTrace ne rend la main qu'une fois que tous les vnements ont t
// traits. Hors pour une trace temps rel a veut dire une fois que la trace
// a t arrte.
// Aussi pour les traces temps rel, on appelle ProcessTrace dans un thread
// spar.
type
  TWorkerThread = class(TThread)
  private
    FSession : TRACEHANDLE; // Handle de la trace  traiter.
  public

    // Cre et initialise le thread. <Client> doit indiquer la classe TETWClient
    // pour laquelle on veut traiter la trace.
    constructor Create(Client : TETWClient);

    // Lorsque le thread est termin le destructeur permet de librer ActiveClient.
    destructor Destroy; override;

    // Dmarre le traitement des vnements de la trace.
    procedure Execute; override;
  end;

// EventCallback est appele par ETW chaque fois qu'un vnement est lu depuis
// la trace. Cette fonction doit lire l'vnement et le traiter.
// En fait, elle sert de relais pour appeler la mthode ProcessEvent de
// l'instance TETWClient qui veut lire la trace. 
procedure EventCallBack(pEvent : PEVENT_TRACE); stdcall;
begin
  // On dlgue simplement la gestion de l'vnement  TETWClient.
  if Assigned(ActiveClient)
  then ActiveClient.ProcessEvent(pEvent);
end;

// Ajoute l'vnement Event dans la trace. Pour les traces temps rels, il
// est possible qu'on flush les buffers dans un ordre diffrent de
// l'ordre chronologique. De sorte qu'on peut recevoir les vnements
// dans le dsordre, mme si ETW les trie avant de nous les envoyer.
// C'est pourqoi AddEvent tri la trace, en faisant attention  insrer
// chaque vnement  la bonne position chronologique.
procedure TETWClient.AddEvent(Event: TEventTrace);
var i : integer;
begin
  Lock; // Obtient un accs exclusif sur la trace. Il ne faut pas qu'un
        // deuxime appel  AddEvent puisse avoir lieu pendant qu'on recherche
        // la position o insrer l'vnement.
  try
    i := FList.Count;

    // Dans un premier temps, on recherche la position d'insertion.
    while (i>0) and (TEventTrace(FList[i-1]).TimeStamp> Event.TimeStamp) do
    begin
      dec(i);
    end;

    if i<0
    then i := 0;
    if i>=FList.Count
    then FList.Add(Event)
    else FList.Insert(i, Event);
  finally
    Unlock; // Libre l'accs exclusif  la trace.
  end;
end;

// Indique que l'on veut modifier la trace. Cette mthode permet d'obtenir
// un acces exclusif  la trace.
procedure TETWClient.Lock;
begin
  WaitForSingleObject(FTraceMutex, 1000);
end;

// Efface les vnements de la trace.
procedure TETWClient.ClearTrace;
var
  i : integer;
begin
  for i := 0 to FList.Count -1 do
  begin
    TObject(FList[i]).Free;
  end;
  FList.Clear;
end;

// Cre et initialise une nouvelle instance de la classe TETWClient :
// Ouvre une trace existante et commence  lire ses vnement. On peut lire
// une trace fichier ou une trace temps rel.
// Si on ouvre une trace fichier, le constructeur rend la main une fois que
// tous les vnements de la trace ont t lus.
// Si on ouvre une trace temps rel, le constructeur rend immdiatement la
// main, les vnements seront lus au fur et  mesure de leur rception en
// tche de fond.
// <Name> : Si on veut ouvrir une trace fichier, <Name> doit contenir le nom
//          du fichier .etl  ouvrir. Si on veut ouvrir une trace temps rel,
//          <Name> doit contenir le nom de la session de trace  ouvrir.
// <IsRealTime> : Indique si on veut ouvrir une trace temps rel (true) ou
//                une trace fichier (false).
// ATTENTION: TETWClient ne peut traiter qu'une seule trace  la fois.
//            Ca ne pose pas de problme avec les traces fichiers puisque
//            le constructeur ne rend la main qu'une fois le traitement de
//            la trace termin.
//            En revanche, si on lit une trace temps rel, on ne pourra pas
//            lire d'autres traces tant que la trace n'aura pas t arrte !
constructor TETWClient.Create(Name: string; IsRealTime: boolean);
var result : cardinal;
begin
  // Mutex pour se proteger des accs simultans  la trace.
  FTraceMutex := CreateMutex(nil, false, nil);
  
  // On ne peut traiter qu'une seule trace  la fois. Il faut donc commencer
  // par vrifier qu'aucune autre trace n'est en cours de traitement.
  if Assigned(ActiveClient)
  then raise Exception.Create('Il y a dj une trace en cours de traitement.');

  // On prpare une structure EVENT_TRACE_LOGFILE pour ouvrir la trace.
  fillchar(FTrace, sizeof(FTrace), 0); // Initialise la structure  0.

  if isRealTime
  then begin
    // Pour les traces temps rel, il faut initialiser LoggerName avec le nom
    // de la trace.
    FTrace.LoggerName := PChar(Name);

    // Le mode de traitement de la trace doit galement tre Temps rel.
    FTrace.LogFileMode := EVENT_TRACE_REAL_TIME_MODE;
  end
  else begin
    // Pour les traces fichiers, il faut indiquer le nom du fichier  traiter
    // dans LogFileName.
    FTrace.LogFileName := PChar(Name);
  end;

  // On dfinit la mthode Callback pour traiter les vnements.
  FTrace.EventCallback := EventCallback;

  // Il ne reste plus qu' ouvrir la trace.
  FSession := OpenTrace(@FTrace);
  if FSession = INVALID_HANDLE_VALUE
  then raiseLastOSError;

  // Pour les traces temps rel, l'information TimeZone n'est pas crite dans
  // L'entte. Il faut la lire aupres du systeme.
  if IsRealTime
  then GetTimeZoneInformation(FTrace.LogfileHeader.TimeZone);

  // Remarque : Dans ETW, il est possible de lire plusieurs fichiers de trace
  //            en appelant plusieurs fois OpenTrace avec des fichiers
  //            diffrents.
  //            Les vnements provenant de chaque fichier sont alors mlangs
  //            et ordonns d'aprs leur timestamp pour tre restitus dans
  //            l'ordre au client.
  //            Cependant notre implmentation ne traite qu'un seul fichier 
  //            la fois.

  FList := TList.Create; // Initialise la liste des vnements de la trace.

  // Maintenant il faut lancer le traitement des vnements. Pour une trace
  // temps, on dlgue cet opration  un thread distinct. Pour une trace
  // fichier on effectue le traitement immdiatement.
  if IsRealTime
  then FWorkerThread := TWorkerThread.create(self)
  else begin
    // On commence par se dfinir comme la trace en cours de traitement. 
    ActiveClient := self;
    try
      // L'appel  de ProcessTrace demande  ETW de commencer  lire les
      // vnements et d'appeler la mthode Callback. 
      result := ProcessTrace(@FSession, // Session  traiter.
                             1,         // Une seule session  traiter.
                             nil,       // Pas de date de dbut.
                             nil);      // Ni de date de fin.
      if (result<>ERROR_SUCCESS) and (result<>ERROR_CANCELLED)
      then raiseLastOSError;
    finally
      // Le traitement de la trace est termin.
      ActiveClient := nil;
    end;
  end;
end;

// Dtruit la classe et libre les ressources.
destructor TETWClient.Destroy;
var i : integer;
begin
  // On s'assure que la variable globale ActiveClient ne plointe plus sur
  // l'instance en cours de destruction.
  if ActiveClient = self
  then ActiveClient := nil;

  // On libre le handle FSession obtenu lors de l'appel  OpenTrace.
  StopTrace;

  // Le mutex sur la trace n'est plus ncessaire.
  CloseHandle(FTraceMutex);
  
  // Il ne reste plus qu' dtruire les donnes de la trace elle mme.
  if Assigned(FList)
  then begin
    for i := 0 to FList.Count-1 do
    begin
      TObject(FList[i]).Free;
    end;
    FList.Free;
  end;
  inherited Destroy;
end;

// Indique que l'on a finit de modifier la trace.
procedure TETWClient.Unlock;
begin
  ReleaseMutex(FTraceMutex);
end;

procedure TETWCLient.StopTrace;
begin
  if FSession<>0
  then begin
    CloseTrace(FSession);
    FSession := 0;
  end;

  if Assigned(FWorkerThread)
  then begin
    FWorkerThread.Free;
    FWorkerThread := nil;
  end;
end;

// Fonction utilitaire permettant de convertir les timestamps de la trace
// en Date et Heure Delphi.
function TETWClient.GetDateTime(ClockTime: int64): TDateTime;
var bias : int64;
begin
  // ClockTime dsigne le nombre de nanoseconde coule depuis le 01/01/1601.
  // Les dates Delphi indiquent le nombre de jour coul depuis le 01/01/1900.
  // Cependant, le temps dans ClockTime est un temps UTC (temps universel).
  // Pour le convertir en date/heure classique, il faut tenir compte du
  // fuseau horaire. Heureusement, ces informations nous sont fournies dans
  // l'entte de la trace : Le dcalage horaire en minutes nous est indiqu
  // dans TimeZone.Bias, selon la formule :
  //  Temps UTC = Temps rel + Bias. 
  bias := FTrace.LogfileHeader.TimeZone.Bias;

  // Il faut lire TimeZone.Bias dans un int64, puis le convertir en Nanosecondes
  // dans un deuxime temps. Si on fait la conversion dans la mme ligne, Delphi
  // fait le calcul avec un integer et donc fait un overflow.
  bias := bias * 60*1000*10000;

  // Il ne reste plus qu' faire la conversion.   
  result := (ClockTime-bias-t0) / TimeDay;
end;

// ProcessEvent est appele par la fonction callback chaque fois qu'un
// vnement de la trace est lu. Cette mthode s'occupe de dcoder
// l'enregistrement et de le mmoriser.
procedure TETWClient.ProcessEvent(pEvent: PEVENT_TRACE);
var
  Data : TEventTrace;
begin
  Data := nil;
  
  // On ne traite que les EventClass SQLEventClass ou DefaultEventClass
  if ((int64(pEvent^.Header.Guid.D1) = int64(SQLEventClass.D1)) and
      (int64(pEvent^.Header.Guid.D4) = int64(SQLEventClass.D4)))
  then Data := TSQLEvent.Create(pEvent);

  if ((int64(pEvent^.Header.Guid.D1) = int64(DefaultEventClass.D1)) and
      (int64(pEvent^.Header.Guid.D4) = int64(DefaultEventClass.D4)))
  then Data := TDefaultEvent.Create(pEvent);

  if Assigned(Data)
  then AddEvent(Data);
end;

// Trie la trace par ordre chronologique.
procedure TETWCLient.SortTrace;
var i,j : integer;
    Event : TEventTrace;
  procedure Swap(a, b : integer);
  var temp : pointer;
  begin
    temp := Trace[a];
    Trace[a] := Trace[b];
    Trace[b] := temp;
  end;
begin
  Lock; // Verrouille la trace en modification pour la dure du trie.
  try
    i := 1;
    while (i<Trace.Count) do
    begin
      Event := TEventTrace(Trace[i]);
      j := i;
      while (j>0) and (TEventTrace(Trace[j-1]).TimeStamp> Event.TimeStamp) do
      begin
        dec(j);
      end;
      if (j>=0) and (i<>j)
      then swap(i, j)
      else inc(i);
    end;
  finally
    Unlock; // Le trie est termin, on dverrouille la trace.
  end;
end;


// Cre et initialise le thread. <Client> doit indiquer la classe TETWClient
// pour laquelle on veut traiter la trace.
constructor TWorkerThread.create(Client : TETWClient);
begin
  // On mmorise le handle de la session pour que la mthode Execute puisse
  // s'en servir.
  FSession := Client.FSession;

  // L'instance ne doit pas tre automatiquement dtruite  la fin du thread.
  FreeOnTerminate := false;

  // Il faut initialiser ActiveClient avant que le thread ne dmarre.
  ActiveClient := Client;

  // Lance l'excution du thread.
  inherited Create(false);
end;

// Lorsque le thread est termin le destructeur permet de librer ActiveClient.
destructor TWorkerThread.Destroy;
begin
  ActiveClient := nil; // Le traitement est termin. On libre ActiveClient.
  inherited Destroy;
end;

// Dmarre le traitement des vnements de la trace.
procedure TWorkerThread.execute;
var r : cardinal;
begin
  try
    r := ProcessTrace(@FSession, 1, nil, nil);
    if (r<>ERROR_SUCCESS) and (r<>ERROR_CANCELLED)
    then raiseLastOSError;
  finally
    ActiveClient := nil;
  end;
end;


end.
