{******************************************************************************}
{                                                                              }
{ Event Tracing for Windows controler Unit                                     }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License          }
{ Version 1.1 (the "License"); you may not use this file except in             }
{ compliance with the License. You may obtain a copy of the License at         }
{ http://www.mozilla.org/MPL/                                                  }
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS"          }
{ basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the      }
{ License for the specific language governing rights and limitations           }
{ under the License.                                                           }
{                                                                              }
{ The Original Code is uETWTraceControler.pas released February 2008.          }
{                                                                              }
{ The Initial Developer of the Original Code is Franck SORIANO.                }
{                                                                              }
{ Alternatively, the contents of this file may be used under the terms of the  }
{ GNU Lesser General Public License (the  "LGPL License"), in which case the   }
{ provisions of the LGPL License are applicable instead of those above.        }
{ If you wish to allow use of your version of this file only under the terms   }
{ of the LGPL License and not to allow others to use your version of this file }
{ under the MPL, indicate your decision by deleting  the provisions above and  }
{ replace  them with the notice and other provisions required by the LGPL      }
{ License.  If you do not delete the provisions above, a recipient may use     }
{ your version of this file under either the MPL or the LGPL License.          }
{                                                                              }
{ For more information about the LGPL: http://www.gnu.org/copyleft/lesser.html }
{                                                                              }
{******************************************************************************}
unit uETWTraceControler;

interface

uses classes, sysutils, windows,
     uETW;

// GUID Utilis pour identifier le controleur de la trace.
const ETWTraceControlerGuid : TGUID = '{6c5a5d8f-c106-4323-80e0-553dcdb510e1}';

type
  // Structure utilise pour dfinir les proprits de la trace
  pTraceProperties = ^TTraceProperties;
  TTraceProperties = packed record
    Props : EVENT_TRACE_PROPERTIES;
    filler : array[0..11] of char;
    SessionName : array[0..511] of char;
    FileName : array[0..511] of char;
  end;
  
  // TETWTraceControler
  // Cette classe reprsente une session de trace active.
  // On dmarre une trace en instanciant la classe. 
  // Lorsque l'instance est dtruite, la trace est automatiquement stoppe.
  TETWTraceControler = class(TObject)
  private
    FRunningSession : TRACEHANDLE;// Handle de la trace en cours.
    FName : string;               // Nom utilis pour crer la trace.
    FFileName : string;           // Nom du fichier de trace si on trace dans un
                                  // fichier.
    FControlGuid : TGUID;         // GUID utilis pour crer la trace.
    FTrcProps : TTraceProperties; // Proprits de la trace en cours.
    FEnabledProviders : TStrings; // Liste des providers activs pour la trace.
  public
    // Cre et initialise une nouvelle instance de la classe TETWTraceControler.
    // Cette opration a pour effet de dmarrer une nouvelle trace dans ETW.
    // <aName> dsigne le nom de la trace. Les noms doivent tre uniques.
    // <aFileName> indique le nom du fichier dans lequel la trace doit tre
    //             crite. Si aucun nom de fichier n'est prcis, une trace
    //             temps rel est lance.
    // Une fois la trace dmarre, il faut activer les providers avec
    // EnableProvider.
    constructor Create(aName : string; aFileName : string);

    // Dtruit la classe et libre les resources associes. La trace est alors
    // arrte.
    destructor Destroy; override;

    // Dsactive le provider <Guid> pour la session de trace. Ce dernier n'crit
    // plus d'vnement dans la trace.
    procedure DisableProvider(Guid : TGUID);

    // Active le provider <Guid> pour la session de trace, avec le niveau de
    // trace <EnableLevel> et les flags <EnableFlags>.
    // Attention, avec ETW, un provider ne peut tre activ que pour une seule
    // trace  la fois. On obtiendra une erreur si le provider est dj actif
    // dans une autre trace.
    procedure EnableProvider(Guid : TGUID; EnableLevel : byte;
      EnableFlags : cardinal);

    // Provoque l'criture immdiate des buffers dans la trace. Pour une trace
    // temps rel, cel permet d'envoyer les vnements au client avant que les
    // buffers ne soient plein.
    // Par contre, il est prfrable de ne pas faire de Flush avec une trace
    // fichier car ce sont les buffers complets qui sont crits sur disque, mme
    // s'ils ne sont pas pleins.
    procedure Flush;

    // Dcode la chane de caractres <provider> pour en extraire le GUID d'un
    // provider ainsi que son tat d'activation.
    // <provider> est une chane retourne par EnumerateRegisteredProviders.
    // <providerGUID> : en sortie, ce paramtre contient le GUID du provider.
    // <Enabled> : En sortie, indique si le provider est actuellement en cours
    //             d'utilisation dans une trace. Un provider ne peut tre activ
    //             que dans une seule trace  la fois.
    class procedure DecodeProviderInfo(provider : string;
                                       var providerGUID : string;
                                       var Enabled : boolean);

    // Retourne dans <aList> la liste des providers disponibles dans ETW, c'est
    //  dire, ceux qui se sont dclars avec  RegisterTraceGuids.
    class procedure EnumerateRegisteredProviders(aList : TStrings);

    // Retourne dans <aList> la liste des traces en cours d'excution sur la
    // machine.
    // La liste est renvoy sous la forme : SessionName + #9 + FileName
    class procedure EnumerateRunningTraces(aList : TStrings);
  end;

implementation
{ TETWTraceControler }
const
  TRACE_ON = 1;
  TRACE_OFF = 0;

// Dcode la chane de caractre <provider> pour en extraire le GUID d'un
// provider ainsi que son tat d'activation.
// <provider> est une chane retourne par EnumerateRegisteredProviders.
// <providerGUID> : en sortie, ce paramtre contient le GUID du provider.
// <Enabled> : En sortie, indique si le provider est actuellement en cours
//             d'utilisation dans une trace. Un provider ne peut tre activ
//             que dans une seule trace  la fois.
class procedure TETWTraceControler.DecodeProviderInfo(provider: string;
  var providerGUID: string; var Enabled: boolean);
var i : integer;
begin
  // provider est dfinit selon le format suivant :
  //     providerGUID=Status
  // pour dcoder la chane, on se contente de la dcouper autour du caractre '='
  i := pos('=', provider);
  if i>0
  then begin // Le sparateur a t trouv, on coupe la chane en deux.
    providerGUID := copy(provider, 1, i-1);
    Enabled := copy(provider, i+1, length(provider)) = '1';
  end
  else begin // Le sparateur n'existe pas. Ce cas ne devrait normalement pas arriver.
    // On considre que la chaine contient uniquement le GUID.
    providerGUID := provider;
    Enabled := false; // Par dfaut, on dit que le provider n'est pas utilis.
  end;
end;

// Dtruit la classe et libre les resources associes. La trace est alors
// arrte.
destructor TETWTraceControler.Destroy;
var i : integer;
begin
  // On commence par dsactiver tous les providers qui ont t activs.
  if Assigned(FEnabledProviders)
  then begin
    // On parcourt la liste  l'envers car DisableProvider va retirer le
    // provider de la liste au fr et  mesure que ces derniers sont dsactivs.
    for i := FEnabledProviders.Count -1 downto 0 do
    begin
      DisableProvider(StringToGuid(FEnabledProviders[i]));
    end;
    FEnabledProviders.Free; 
  end;

  StopTrace(0, PChar(FName), @FTrcProps);
{  // Arrte la trace en lui envoyant l'vnement EVENT_TRACE_CONTROL_STOP
  // FTRcProps a t initialis au moment de la cration de la trace.
  ControlTrace(FRunningSession, PChar(FName), @FTrcProps, EVENT_TRACE_CONTROL_STOP);}
end;

// Dsactive le provider <Guid> pour la session de trace. Ce dernier n'crit
// plus d'vnement dans la trace.
procedure TETWTraceControler.DisableProvider(Guid: TGUID);
var idx : integer;
begin
  // Arrt du provider dans ETW.
  if EnableTrace(TRACE_OFF, 0, 0, @GUID, FRunningSession)<>ERROR_SUCCESS
  then RaiseLastOSError;

  // On doit  prsent retirer le provider de la liste des providers actifs.
  idx := FEnabledProviders.IndexOf(GuidToString(Guid));
  if idx<>-1
  then FEnabledProviders.Delete(idx);
end;

// Active le provider <Guid> pour la session de trace, avec le niveau de trace
// <EnableLevel> et les flags <EnableFlags>.
procedure TETWTraceControler.EnableProvider(Guid : TGUID; EnableLevel : byte; EnableFlags : cardinal);
begin
  // On commence par vrifier si le provider n'a pas dj t activ.
  if FEnabledProviders.IndexOf(GuidToString(Guid)) = -1
  then begin
    // Activation du provider.
    if EnableTrace(TRACE_ON, EnableFlags, EnableLevel,
                    @GUID, FRunningSession)<>ERROR_SUCCESS
    then RaiseLastOSError;

    // On ajoute le provider  la liste des providers dj activs.
    FEnabledProviders.Add(GuidToString(Guid));
  end;
end;

// Retourne dans <aList> la liste des providers disponibles dans ETW, c'est
//  dire, ceux qui se sont dclars avec  RegisterTraceGuids.
class procedure TETWTraceControler.EnumerateRegisteredProviders(
  aList: TStrings);
var nbProviders : cardinal;
    Properties : array of TRACE_GUID_PROPERTIES;
    PropertiesPtr : array of PTRACE_GUID_PROPERTIES;
    MaxProv : cardinal;
    i : integer;
    result : cardinal;
    s : string;

  // Initialise les tableaux <Properties> et <PropertiesPtr> pour stocker un
  // maximum de <nbProv>.
  procedure InitTabs(nbProv : integer);
  var i : integer;
  begin
    // Properties est un tableau linaire de structures TRACE_GUID_PROPERTIES.
    // PropertiesPtr est un tableau qui pointe sur les lments de Properties
    // un  un.
    SetLength(Properties, nbProv);
    SetLength(PropertiesPtr, nbProv);

    // On commence par initialiser les tableaux  0.
    fillchar(Properties[0], sizeof(TRACE_GUID_PROPERTIES) * nbProv, 0);

    // Il ne reste plus qu' initialiser chaque lment de PropertiesPtr pour les
    // faire pointer sur un lment de Properties.
    for i := 0 to nbProv -1 do
    begin
      PropertiesPtr[i] := @Properties[i];
    end;
  end;

begin
  Assert(Assigned(aList)); // aList doit avoir t initialis cr par
                           // l'appelant.
  aList.Clear;

  // Par dfaut, on prvoie de lire 50 providers.
  MaxProv := 50;
  nbProviders := MaxProv;

  // On prpare un tableau pour MaxProv fournisseurs au maximum.
  InitTabs(MaxProv);

  // On demande  ETW d'numrer les providers dclars.
  // EnumerateTraceGuids va remplir le tableau Properties avec la liste des
  // providers disponibles. <MaxProv> indique la taille maximal du tableau.
  // En retour <nbProviders> contiendra le nombre total de provider dclar dans
  // ETW.
  // Si le tableau Properties est trop petit (il y a plus 50 providers dclar)
  // EnumerateTraceGuids renverra le code de retour ERROR_MORE_DATA. Dans ce cas
  // il faut recommencer l'opration avec un tableau plus grand.
  result := EnumerateTraceGuids(PropertiesPtr, MaxProv, nbProviders);
  case result of
  ERROR_SUCCESS:; // Tout est Ok.
  ERROR_MORE_DATA: // Il y a plus de provider que prvu, il faut agrandir le
                   // tableau.
    begin
      MaxProv := nbProviders;
      InitTabs(MaxProv); // On redfinit les tableaux. Cette fois on connait le
                         // nombre exacte de provider qui seront renvoy.

      // Il ne reste plus qu' refaire la lecture.                         
      if EnumerateTraceGuids(PropertiesPtr, MaxProv, nbProviders)<>ERROR_SUCCESS
      then raiseLastOSError;
    end;
  else RaiseLastOSError;
  end;

  // Les informations ont t obtenues dans Properties. On met le rsultat en
  // forme pour remplir <aList>.
  for i := 0 to nbProviders-1 do
  begin
    s := GuidToString(Properties[i].Guid);
    if Properties[i].IsEnable<>0
    then s := s + '=1'
    else s := s + '=0';
    aList.Add(s);
  end;
end;

// Retourne dans <aList> la liste des traces en cours d'excution sur la
// machine.
// La liste est renvoy sous la forme : SessionName + #9 + FileName
class procedure TETWTraceControler.EnumerateRunningTraces(aList: TStrings);
var nbTraces : cardinal;
    TrcProps : array of TTraceProperties;
    TrcPropsPtr : array of PTraceProperties;
    MaxTraces : cardinal;
    i : integer;
    result : cardinal;
    s : string;

  // Initialise les tableaux <TrcProps> et <TrcPropsPtr> pour stocker un
  // maximum de <nbTrc> traces.
  procedure InitTabs(nbTrc : integer);
  var i : integer;
  begin
    // TrcProps est un tableau linaire de structures TTraceProperties.
    // TrcPropsPtr est un tableau qui pointe sur les lments de TrcProps
    // un  un.
    SetLength(TrcProps, nbTrc);
    SetLength(TrcPropsPtr, nbTrc);

    // On commence par initialiser les tableaux  0.
    fillchar(TrcProps[0], sizeof(TTraceProperties) * nbTrc, 0);
    
    for i := 0 to nbTrc -1 do
    begin
      TrcPropsPtr[i] := @TrcProps[i];

      // Il faut initialiser BufferSize pour indiquer la taille de la structure.  
      TrcProps[i].Props.Wnode.BufferSize := sizeof(TrcProps[i]);
      TrcProps[i].Props.LoggerNameOffset := 128;
      TrcProps[i].Props.LogFileNameOffset := 256;       
    end;
  end;

begin
  assert(Assigned(aList)); // aList doit avoir t initialis cr par l'appelant.
  aList.Clear;

  // Par dfaut, on prvoit lire un maximum de 32 traces.
  MaxTraces := 32;
  nbTraces := 0;

  // Initialisation du tableau de structures.
  InitTabs(MaxTraces);

  // On lit la liste des traces avec QueryAllTraces. S'il y a plus de <MaxTraces>
  // traces actives, QueryAllTraces va renvoyer ERROR_MORE_DATA. Dans ce cas
  // il faut agrnadir les tableaux et recommencer.
  nbTraces := MaxTraces;
  result := QueryAllTraces(TrcPropsPtr, MaxTraces, nbTraces);
  case result of
  ERROR_SUCCESS:; // Tout est OK.
  ERROR_MORE_DATA: // Il y a plus de traces  lire.
    begin
      // On agrandit le nombre maximum de trace. Cette fois, on connait le nombre
      // exacte.
      MaxTraces := nbTraces;
      InitTabs(MaxTraces);

      // On recommence la lecture.
      if QueryAllTraces(TrcPropsPtr, MaxTraces, nbTraces)<>ERROR_SUCCESS
      then RaiseLastOSError;
    end;
  else RaiseLastOSError;
  end;

  // Il ne reste plus qu' mettre le rsultat en forme.
  for i := 0 to nbTraces -1 do
  begin
    s := StrPas(PChar(integer(@TrcProps[i]) + TrcProps[i].Props.LoggerNameOffset));
    s := s + #9 + StrPas(PChar(integer(@TrcProps[i]) + TrcProps[i].Props.LogFileNameOffset));
    aList.Add(s);
  end;
end;

// Provoque l'criture immdiate des buffers dans la trace. Pour une trace temps
// rel, cel permet d'envoyer les vnements au client avant que les buffers
// ne soient plein.
// Par contre, il est prfrable de ne pas faire de Flush avec une trace
// fichier car ce sont les buffers complets qui sont crits sur disque, mme
// s'ils ne sont pas pleins.
procedure TETWTraceControler.Flush;
begin
  // On demande que les buffers soient immdiatement crits
  ControlTrace(FRunningSession, nil, @FTrcProps, EVENT_TRACE_CONTROL_FLUSH);
end;


// Cre et initialise une nouvelle instance de la classe TETWTraceControler.
// Cette opration a pour effet de crer une nouvelle session de trace dans
// ETW, et de dmarrer la trace.
// <aName> dsigne le nom de la trace. Les noms doivent tre uniques.
// <aFileName> indique le nom du fichier dans lequel la trace doit tre crite.
//             si aucun nom de fichier n'est prcis, une trace temps rel
//             est lance.
// Une fois la trace dmarre, il faut activer les providers avec EnableProvider.
constructor TETWTraceControler.Create(aName: string; aFileName : string);
begin
  // On mmorise les paramtres d'appel.
  FName := aName;
  FFileName := aFileName;
  FControlGuid := ETWTraceControlerGuid;

  // Il faut toujours commencer par initialiser la structure  0.
  fillchar(FTrcProps, sizeof(FTrcProps), 0);
  FTrcProps.Props.Wnode.BufferSize := sizeof(FTrcProps); // On dfinit la taille
                                                         // de la structure.
  FTrcProps.Props.Wnode.Flags := WNODE_FLAG_TRACED_GUID; // Valeur obligatoire.
  FTrcProps.Props.Wnode.ClientContext := 1; // Les TimeStamp seront dfinis
                                            // avec les compteurs de performance
                                            // windows.
  FTrcProps.Props.Wnode.Guid := FControlGuid; // Identifiant de la session.
  FTrcProps.Props.BufferSize := 64; // Taille des buffers en Ko.
  FTrcProps.Props.MinimumBuffers := 3; // Nombre de buffer  prallouer.
  FTrcProps.Props.MaximumBuffers := 16; // Nombre maximum de buffers.

  // LoggerNameOffset doit tre initialis avec l'offset de SessionName
  // par rapport  la structure EVENT_TRACE_PROPERTIES. Ca aurait d tre
  // sizeof(FTrcProps.Props). Cependant, avec cette valeur, StartTrace renvoie
  // toujours ERROR_INVALID_PARAMETER.
  // Par contre une valeur de 128 (> la position normale) est accepte.
  // Il s'agit peut tre d'un problme d'alignement des adresses en mmoire.
  // C'est pourquoi j'ai ajout un champ Filler pour aligner SessionName sur
  // l'offset 128.
  FTrcProps.Props.LoggerNameOffset := sizeof(FTrcProps.Props) +
                                      sizeof(FTrcProps.filler);
  FTrcProps.Props.LogFileNameOffset := FTrcProps.Props.LoggerNameOffset +
                                       sizeof(FTrcProps.SessionName);

  // On dfinit le nom du fichier de trace.
  // Attention, pour TETWTraceControler, un nom de fichier vide signifie une trace
  // temps rel. Cependant avec ETW, il est possible de dfinir une trace temps
  // rel et un nom de fichier. En principe, ETW envoie alors les vnements au
  // client en temps rel, tout en les crivant sur disque.
  strcopy(@FTrcProps.FileName[0], PChar(FFileName));

  if aFileName = '' // Pas de nom de fichier, on cre une trace temps rel.
  then begin
    FTrcProps.Props.LogFileMode := EVENT_TRACE_REAL_TIME_MODE or
                                   EVENT_TRACE_USE_PAGED_MEMORY;
    FTrcProps.Props.LogFileNameOffset := FTrcProps.Props.LoggerNameOffset +
        sizeof(FTrcProps.SessionName);

    FTrcProps.Props.FlushTimer := 1; // Les buffers doivent tre crits toutes
                                     // les secondes.
  end
  else begin
    // On a un nom de fichier, on cre une trace fichier.
    FTrcProps.Props.LogFileMode := EVENT_TRACE_FILE_MODE_CIRCULAR or
                                   EVENT_TRACE_USE_PAGED_MEMORY;
    FTrcProps.Props.MaximumFileSize := 100;
  end;

  // Commence par essayer d'arrter la trace pour le cas o il existerait dj
  // une trace avec le mme nom. C'est par exemple le cas si on a dmarr une
  // trace temps rel et que le contrleur n'a pas t ferm proprement.  
  StopTrace(0, PChar(FName), @FTrcProps);

  // On demande  ETW de dmarrer la trace.
  if StartTrace(FRunningSession, PChar(FName), @FTrcProps)<>ERROR_SUCCESS
  then raiseLastOSError;

  FEnabledProviders := TStringList.Create;
end;

end.
