{******************************************************************************}
{                                                                              }
{ Event Tracing for Windows WMI MOF decoder Unit                               }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License          }
{ Version 1.1 (the "License"); you may not use this file except in             }
{ compliance with the License. You may obtain a copy of the License at         }
{ http://www.mozilla.org/MPL/                                                  }
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS"          }
{ basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the      }
{ License for the specific language governing rights and limitations           }
{ under the License.                                                           }
{                                                                              }
{ The Original Code is uETWWmiInfo.pas released February 2008.                 }
{                                                                              }
{ The Initial Developer of the Original Code is Franck SORIANO.                }
{                                                                              }
{ Alternatively, the contents of this file may be used under the terms of the  }
{ GNU Lesser General Public License (the  "LGPL License"), in which case the   }
{ provisions of the LGPL License are applicable instead of those above.        }
{ If you wish to allow use of your version of this file only under the terms   }
{ of the LGPL License and not to allow others to use your version of this file }
{ under the MPL, indicate your decision by deleting  the provisions above and  }
{ replace  them with the notice and other provisions required by the LGPL      }
{ License.  If you do not delete the provisions above, a recipient may use     }
{ your version of this file under either the MPL or the LGPL License.          }
{                                                                              }
{ For more information about the LGPL: http://www.gnu.org/copyleft/lesser.html }
{                                                                              }
{******************************************************************************}
unit uETWWmiInfo;

interface
uses classes, sysutils, windows, WbemScripting_TLB;

type
  TWmiObject=class;

  // TETWWmiInfo : Cette classe est une faade permettant de lire le schma
  // des vnements publis dans Wmi avec MOF.
  // Elle est gre en singleton par l'intermdiaire de la fonction WmiInfo.
  TETWWMIInfo = class(TObject)
  private
    FClassIndex : TStrings; // Liste des EventClass. 
    FEventIndex : TStrings; // Index permettant de retrouver rapidement la
                            // dfinition d'un vnement ETW
    FProviders : TList;     // Index des providers ETW trouvs.

    // Cre et initialise une nouvelle instance de la classe TETWWMIInfo.
    // Lorsque la classe est instancie, on se connecte automatiquement  Wmi
    // pour numre toutes les informations sur les providers ETW dclars.
    // Une fois l'instanciation termine, la classe n'accde plus  Wmi.
    constructor Create;    
  public
    // Dtruit l'instance et libre les ressources.
    destructor Destroy; override;

    // Renvoie le nom publique dfinit pour le provider identifi  partir du
    // GUID <providerGUID>. Si la mthode ne parvient pas  trouver de nom,
    // elle retourne <providerGUID> sous la forme d'une chane de caractres.
    function GetProviderName(providerGUID : string) : string;

    // Recherche la classe TWmiObject qui dcrit l'vnement ETW identifi par
    // l'EventClass <Guid>, la version <EventVersion> et le type d'vnement
    // <EventType>.
    // Si aucune information n'a t dfinie pour cet vnement, la mthode
    // renvoie nil.
    function GetEventInfo(Guid : TGUID; EventVersion : word;
      EventType : byte) : TWmiObject;

    // collection d'instance TWmiObject reprsentant les providers ETW dclars
    // dans Wmi.
    property Providers : TList read FProviders;
  end;

  // TWmiObject : Cette classe reprsente une classe WMI.
  // Il ne faut pas crer d'instance de cette classe directement. A la place,
  // il faut plutt utiliser TETWWMIInfo qui s'occupe de charger toutes les
  // informations relatives au schma des vnements ETW.
  TWmiObject = class(TObject)
  private
    FWmiObject : ISWbemObject; 
    FGuid: string;
    FChilds : TList;
    FProperties : TStrings;
    FDescription : string;
    FVersion : integer;
    EventTypes : array of integer;
    EventTypeNames : array of string;
    FClassName2 : string;

    function GetDescription: string;
    function GetClassName2: string;

    // Cre et initialise une nouvelle instance de la classe TWmiObject  partir
    // des informations lues depuis Wmi.
    // <WmiObject> : Interface sur la classe Wmi modlise.
    // <Service> : Interface sur le service Wmi.
    // <ClassIndex> : Dictionnaire mmorisant la liste des vnements ETW
    //                trouvs. Le dictionnaire est remplit au fur et  mesure
    //                des classes numres.
    // <Level> : indique la profondeur courante de la classe dans l'arbre
    //           d'hritage wmi. 1 pour les descendants direct de EventTrace
    //           2 pour les classes drivant d'un provider ETW et 3 pour les
    //           types d'vnemenets Wmi.
    constructor Create(WmiObject : ISWbemObject;
                       Service : ISWbemServices;
                       ClassIndex : TStrings;
                       Level : integer);
  public
    // Dtruit l'instance ainsi que tous ses enfants.
    destructor Destroy; override;

    // Dcode les donnes utilisateur d'un vnement ETW et les met en forme
    // dans une chane de caractres.
    // <MofData> doit pointer sur les donnes utilisateur de l'vnement.
    // <MofLength> doit indiquer la taille total des donnes pointes par
    // <MofData>.
    function GetUser(MofData : pointer; MofLength : integer) : string;

    // Retourne le nom associ au type d'vnement <EventType>. Si aucun nom
    // n'a t dfini, GetEventTypeName retourne EventType sous la forme d'une
    // chane de caractres.
    function GetEventTypeName(EventType : integer) : string;

    // Guid associ  la classe Wmi, si ce dernier tait dfinit.
    property Guid : string read FGuid;

    // Description dfinie pour l'vnement.
    property Description : string read FDescription write FDescription;

    // Nom de la classe Wmi reprsente par la classe.
    property ClassName2 : string read GetClassName2;

    // Liste des classes drives. Si l'instance reprsente un provider ETW,
    // Childs contient la liste des EventClass dfinis pour ce provider.
    // Si l'instance reprsente un EventClass, Childs contient la liste des
    // EventType dfinit pour l'vnement.
    // Si l'instance reprsente un EventType, Childs est vide.
    property Childs : TList read FChilds;

    // Liste des proprits de la classe Wmi. Cette liste n'est remplie que si
    // l'instance reprsente un EventType.
    property Properties : TStrings read FProperties;

    // Version associe au EventClass. Si l'instance ne dsigne pas un
    // EventClass, version vaut 0.
    property Version : integer read FVersion write FVersion;
  end;

  // TMofField reprsente un champ d'un vnement ETW. La classe permet de
  // dcoder un champ depuis le buffer MofData, et de le formater selon les
  // indications dclares dans la classe Wmi.
  TMofField = class(TObject)
  private
    FPropertyName: string;
    FDisplayLabel: string;
    FWmiDataId: integer;
    FDataType: integer;
    FIsArray : boolean;
    FIsPointer : boolean;
    FArraySize : integer;
    FExtension : string;
    FFormat : string;
    FStringTermination : string;
    FValueType : string;
    FXMLFragment : boolean;

    function ArrayResult(Valeur : string) : string;

    // Formate une donne de type Integer.
    function FormatSInt32(var Offset: Cardinal; MofData: Pointer) : string;

    // Formate une donne de type Cardinal.
    function FormatUInt32(var Offset: Cardinal; MofData: pointer) : string;

    // Formate une donne de type int64.
    function FormatInt64(var Offset: Cardinal; MofData: Pointer) : string;

    // Formate une donne de type string.
    function FormatString(var Offset: Cardinal; MofData: Pointer;
      MofLength : cardinal) : string;

    // Formate une donne de type boolean
    function FormatBoolean(var Offset: Cardinal; MofData: Pointer) : string;

    // Formate une donne de type shortint.
    function FormatSint8(var Offset: Cardinal; MofData: Pointer) : string;

    // Formate une donne de type byte.
    function FormatUInt8(var Offset: Cardinal; MofData: Pointer) : string;

    // Formate une donne de type word.
    function FormatSInt16(var Offset: Cardinal; MofData: Pointer) : string;

    // Formate une donne de type Objet.
    function FormatObject(var Offset: Cardinal; MofData: Pointer) : string;

    // Formate une donne de type word.
    function FormatUInt16(var Offset: Cardinal; MofData: Pointer) : string;

  public
    BitMap : array of integer;
    BitValues : array of string;
    ValueMap : array of cardinal;
    Values : array of string;

    // Cre et initialise une nouvelle instance de la claase TMofField
    constructor create(prop : ISWbemProperty);

    // Formate la valeur passe en paramtre en fonction des options de
    // formatage dfinies pour la proprit. En particulier, remplace les
    // valeurs de flag par les constantes associes, formate les types
    // numrs...
    function FormatVal(valeur : cardinal) : string;

    // Nom de la proprit Wmi, c'est  dire le bom du champ.    
    property PropertyName : string read FPropertyName write FPropertyName;

    // Description du champ.
    property DisplayLabel : string read FDisplayLabel write FDisplayLabel;

    // Numro d'ordre du champ dans l'vnement.
    property WmiDataId : integer read FWmiDataId write FWmiDataId;

    // Type de donne contenu dans le champ. Il s'agit d'une valeur CIM_TYPE.
    property DataType : integer read FDataType write FDataType;

    // Indique si le champ est un tableau de valeurs.
    property IsArray : boolean read FIsArray write FIsArray;

    // Indique si le champ est un pointeur sur la valeur spcifie.
    property IsPointer : boolean read FIsPointer write FIsPointer;

    // Dans le cas d'un tableau, indique le nombre d'lment du tableau. S'il
    // ne s'agit pas d'un tableau, ArraySize vaut 1.
    property ArraySize : integer read FArraySize write FArraySize;

    // Valeur de l'extension dfinie pour la proprit.
    property Extension : string read FExtension write FExtension;

    // Valeur du Format dfinit pour la proprit.
    property FormatQualifer : string read FFormat write FFormat;

    // Indique la faon dont la longueur des chaines de caractres est dfinie.
    property StringTermination : string read FStringTermination
      write FStringTermination;

    // Indique la faon dont il faut interprter la valeur si ValueMap et Values
    // sont dfinis (flag ou index).
    property ValueType : string read FValueType write FValueType;

    // Indique si le champ contient un fragment XML.
    property XMLFragment : boolean read FXMLFragment write FXMLFragment;

    // Extrait et formate la valeur du champ depuis le buffer de l'vnement
    // <MofData>.
    // <MofLength> indique la taille total du buffer point par <MofData>.
    // En entre, <Offset> doit dsigner la position du champ par rapport a
    // <MofData>. En sortie, <Offset> contient la position du prochaine champ
    // dans <MofData>.
    function GetValue(MofData : pointer; MofLength : cardinal;
      var Offset : cardinal) : string; virtual;
  end;

// WmiInfo retourne l'instance unique de la classe TETWWmiInfo gre en singleton.
function WmiInfo : TETWWMIInfo;

// FormatHex est une fonction utilitaire qui permet de dumper le buffer <MofData>
// sous la forme d'une chane de caractres hexadcimale.
// <MofLength> doit dsigner la taille du buffer point par <MofData>.
function FormatHex(MofData : pointer; MofLength : integer) : string;

implementation
uses variants, ActiveX;

var UniqueInstance : TETWWMIInfo = nil;

const SECURITY_MAX_SID_SIZE = 68;

// WmiInfo retourne l'instance unique de la classe TETWWmiInfo gre en singleton.
function WmiInfo : TETWWMIInfo;
begin
  if Not Assigned(UniqueInstance)
  then UniqueInstance := TETWWMIInfo.Create;

  result := UniqueInstance;
end;

{ TETWWMIInfo }

// Cre et initialise une nouvelle instance de la classe TETWWMIInfo.
// Lorsque la classe est instancie, on se connecte automatiquement  Wmi
// pour numre toutes les informations sur les providers ETW dclars.
// Une fois l'instanciation termine, la classe n'accde plus  Wmi.
constructor TETWWMIInfo.Create;
var
  Locator : ISWbemLocator;
  Service : ISWbemServices;
  objList : ISWbemObjectSet;
  WmiObj, WmiObj2 : TWmiObject;
  Enum : IEnumVariant;
  Fetch : cardinal;
  TempObj : OleVariant;
  i,j,k : integer;
  Key : string;
  version : word;

begin
  FProviders := TList.Create;
  FClassIndex := TStringList.Create;
  TStringList(FClassIndex).Sorted := true;
  TStringList(FClassIndex).Duplicates := dupAccept;

  FEventIndex := TStringList.Create;
  TStringList(FEventIndex).Sorted := true;  

  // Dbut de la lecture des informations WMI.
  Locator := CoSWbemLocator.Create; // Instanciation de l'objet COM.

  // On se connecte au namespace root\wmi du serveur local.
  Service := Locator.ConnectServer('', 'root\wmi', '', '', '', '', 0, nil);

  // On numre toutes les classes Wmi qui drivent de la classe EventTrace
  // pour obtenir la liste des providers ETW.
  objList := Service.SubclassesOf('EventTrace',
                wbemQueryFlagShallow, // On ne veut que les descendant direct.
                nil);
  // objList contient la liste de rsultat. Pour la parcourir, il faut passer
  // un numrateur avec l'interface IEnumVariant (cette dernire permet
  // d'implmenter les foreach de VB.
  Enum := (objList._NewEnum) as IEnumVariant;
  while Enum.Next(1, TempObj, Fetch) = S_OK do
  begin
    // On a trouv une classe reprsentant un provider. On construit un objet
    // TWmiObject pour mmoriser les informations du provider. Ce dernier va
    // lire recursivement les informations des EventClass et EventType.
    WmiObj := TWmiObject.Create(IUnknown(TempObj) as ISWbemObject,
                  Service, // Pour faire les numrations rcursivement
                  FClassIndex, // Index des objets TWmiObject construits.
                  1); // On commence au niveau 1, c'est  dire il s'agit d'un
                      // provdier.

    if (WmiObj.Guid = '') or // Si la classe Wmi ne possde pas de Guid, il ne
                             // s'agit pas d'un provider. 
       (FClassIndex.IndexOf(WmiObj.Guid)=-1) 
    then WmiObj.Free // Si l'instance n'a pas t mmorise dans FClassIndex,
                     // on la dtruit car elle ne sera pas utilise.
    else FProviders.Add(WmiObj); // Sinon on l'ajoute  la liste des providers.
  end;

  // A prsent, toutes les classes Wmi utilent ont t charges et une instance
  // TWmiObject a t cre et stocke dans FClassIndex.
  // On va construire un index permettant d'identifier immdiatement la classe
  //  utiliser pour dcoder un vnement ETW en fonction de son EventClass,
  // de sa version et de son EventType.

  // Pour cel, on parcours la liste des classes trouves et on construit un
  // index dans FEventIndex.
  for i := 0 to FClassIndex.Count-1 do
  begin
    // FClassIndex est dj un dictionnaire dont la cl n'est compose que du
    // Guid de l'EventClass.
    Key := FClassIndex[i]; // Le guid constitue le dbut de la cl.
    WmiObj := TWmiObject(FCLassIndex.Objects[i]); 
    if Assigned(WmiObj)
    then begin
      // On intgre le numro de version de la classe EventClass  la cl de
      // l'index.
      Version := WmiObj.Version;
      Key := Key + '|' + IntToStr(Version);

      // A prsent, il faut parcourir les classes enfants pour indexer les
      // EventType.
      for j := 0 to WmiObj.Childs.Count -1 do
      begin
        WmiObj2 := WmiObj.Childs[j];
        if Assigned(WmiObj2)
        then begin
          // On en profite pour initialiser la description de l'EventType avec
          // celle du EventClass. En effet, c'est elle qui donne le nom de
          // l'vnement. De cette faon, la classe de l'EventType contiendra
          // immdiatement toutes les informations ncessaires pour formater
          // un vnement. 
          WmiObj2.Description := WmiObj.Description;

          // On construit une entre dans l'index par EventType support par
          // la classe. Chaque entre correspondra  la mme classe bien
          // videmment.
          for k := low(wmiObj2.EventTypes) to high(wmiObj2.EventTypes) do
          begin
            // Ajout de l'entre au dictionnaire.
            FEventIndex.addObject(Key + '|' + IntToStr(wmiObj2.EventTypes[k]),
              wmiObj2);
          end;
          WmiObj2.FWmiObject := nil; // Enfin, on n'a plus besoin de garder la
                                     // rfrence sur l'objet COM WMI.
        end;
      end;
      WmiObj.FWmiObject := nil; 
    end;
  end;

  Service := nil; // On libre la rfrence sur l'objet COM.
  Locator := nil; // On libre la rfrence sur l'objet COM.
end;

// Dtruit l'instance et libre les ressources.
destructor TETWWMIInfo.Destroy;
var i : integer;
begin
  // FProviders est un arbre qui contient toutes les instances TWmiObject qui
  // ont t cres. Il faut dtruire ces instances en mme temps que FProviders.
  for i := 0 to FProviders.Count -1 do
  begin
    TObject(FProviders[i]).free;
  end;
  FProviders.free;

  // Par contre FClassIndex et FEventIndex ne sont que des indexes qui rfrences
  // les instances de FProviders. Il faut simplement dtruire les listes
  // sans chercher  dtruire les objets rfrencs.
  FClassIndex.free;
  FEventIndex.Free;
  inherited Destroy; 
end;

// Recherche la classe TWmiObject qui dcrit l'vnement ETW identifi par
// le EventClass <Guid>, la version <EventVersion> et le type d'vnement
// <EventType>.
// Si aucune information n'a t dfinie pour cet vnement, la mthode
// renvoie nil.
function TETWWMIInfo.GetEventInfo(Guid: TGUID; EventVersion: word;
  EventType: byte): TWmiObject;
var idx : integer;

begin
  // Il suffit de rechercher la classe dans l'index.
  idx := FEventIndex.IndexOf(GuidToString(Guid) + '|' + IntToStr(EventVersion) + '|' + IntToStr(EventType));
  if idx<>-1
  then result := TWmiObject(FEventIndex.Objects[idx])
  else result := nil; // Aucune classe trouve, on renvoie nil.
end;

// Renvoie le nom publique dfinit pour le provider identifi  partir du
// GUID <providerGUID>. Si la mthode ne parvient pas  trouver de nom,
// elle retourne <providerGUID> sous la forme d'une chane de caractres.
function TETWWMIInfo.GetProviderName(providerGUID: string): string;
var idx : integer;
    WmiObj : TWmiObject;
begin
  // On recherche le guid du provider dans l'index FClassIndex.
  idx := FClassIndex.IndexOf(providerGUID);
  if idx<>-1
  then begin
    WmiObj := TWmiObject(FClassIndex.Objects[idx]);
    result := WmiObj.Description;
  end
  else result := providerGUID; // Aucune info disponible, on retourne le GUID
end;


{ TWmiObject }

// Cre et initialise une nouvelle instance de la classe TWmiObject  partir
// des informations lues depuis Wmi.
// <WmiObject> : Interface sur la classe Wmi modlise.
// <Service> : Interface sur le service Wmi.
// <ClassIndex> : Dictionnaire mmorisant la liste des vnements ETW
//                trouvs. Le dictionnaire est remplit au fur et  mesure
//                des classes numres.
// <Level> : indique la profondeur courante de la classe dans l'arbre
//           d'hritage wmi. 1 pour les descendants direct de EventTrace
//           2 pour les classes drivant d'un provider ETW et 3 pour les
//           types d'vnemenets Wmi.
constructor TWmiObject.Create(WmiObject: ISWbemObject;Service : ISWbemServices;
    ClassIndex : TStrings; Level : integer);
var
  objList : ISWbemObjectSet;
  Enum : IEnumVARIANT;
  tempObj : OleVariant;
  Fetch : cardinal;
  Child : TWmiObject;
  prop : ISWbemProperty;
  Field : TMofField;
  v : OleVariant;
  k : integer;
  FWmiDataId : cardinal;
begin
  FChilds := TList.Create;
  FProperties := TStringList.Create;
  TStringList(FProperties).Sorted := true;

  FWmiObject := WmiObject;

  // Lecture du GUID associ  la classe Wmi.
  try
    FGuid := FWmiObject.Qualifiers_.Item('Guid', 0).Get_Value;
    ClassIndex.AddObject(FGuid, self);
  except
    FGuid := '';
  end;

  // Lecture de la description
  FDescription := GetDescription;

  // Lecture du nom de la classe WMI
  FClassName2 := GetClassName2;

  // Si la classe dsigne un EventClass, il faut lire son EventVersion.
  if Level = 2
  then begin
    try
      Version := FWmiObject.Qualifiers_.Item('EventVersion', 0).Get_Value;
    except
      Version := 0;
    end;
  end
  else Version := 0; // Sinon on affecte 0 comme Version par dfaut.

  // Si la classe est un EventType, il faut lire ses proprits.
  if Level = 3
  then begin
    // D'abord on dcode son EventType 
    try
      v := FWmiObject.Qualifiers_.Item('EventType', 0).Get_Value;
      if VarisArray(v) // On peut avoir un tableau de valeurs.
      then begin
        // Dans ce cas, on recopie le tableau dans un tableau Delphi classique.
        SetLength(EventTypes, VarArrayHighBound(v, 1)+1);
        for k := 0 to VarArrayHighBound(v, 1) do
        begin
          EventTypes[k] := VarArrayGet(v, [k]);
        end;
      end
      else begin
        // Si on a directement une valeur, on construit qu'en mme un tableau
        // d'un seul lment. 
        SetLength(EventTypes, 1);
        EventTypes[0] := v;
      end;
    except
      SetLength(EventTypes, 1);
      EventTypes[0] := v;
    end;

    // On dcode galement l'EventTypeName correspondant.
    try
      v := FWmiObject.Qualifiers_.Item('EventTypeName', 0).Get_Value;
      if VarisArray(v) // On peut avoir un tableau de valeur.
      then begin
        // Dans ce cas, on recopie le tableau dans un tableau Delphi classique.      
        SetLength(EventTypeNames, VarArrayHighBound(v, 1)+1);
        for k := 0 to VarArrayHighBound(v, 1) do
        begin
          EventTypeNames[k] := VarArrayGet(v, [k]);
        end;
      end
      else begin
        // Si on a directement une valeur, on construit qu'en mme un tableau
        // d'un seul lment.
        SetLength(EventTypeNames, 1);
        EventTypeNames[0] := v;
      end;
    except
    end;

    // Enfin, on lit les proprits.
    Enum := FWmiObject.Properties_._NewEnum as IEnumVARIANT;
    FWmiDataId := 0;    
    while Enum.Next(1, TempObj, Fetch) = S_OK do
    begin
      prop := IUnknown(TempObj) as ISWbemProperty;
      try
        FWmiDataId := prop.Qualifiers_.Item('WmiDataId', 0).Get_Value;
      except
        continue; // On ne traite que les proprits qui sont marques avec
                  // WmiDataId.
      end;
      Field := TMofField.create(prop);
      Properties.AddObject(Format('%0.3d', [FWmiDataId]), Field);
    end;
  end;

  // Enfin, si on n'est pas dj au niveau des EventType, on continue les
  // numration rcursivement.
  if Level < 3
  then begin
    // On recherche toutes les classes qui drivent de la classe courante.
    objList := Service.SubclassesOf(ClassName2,
                  wbemQueryFlagShallow, // On ne veut que les descendants directs.
                  nil);
    Enum := (objList._NewEnum) as IEnumVariant;
    while Enum.Next(1, TempObj, Fetch) = S_OK do
    begin
      // Pour chaque classe trouve, on construit une instance TWmiObject.
      Child := TWmiObject.Create(IUnknown(TempObj) as ISWbemObject, Service,
          ClassIndex,
          Level+1);
      FChilds.Add(Child); // Et on l'ajoute  la liste des enfants.
    end;
  end;

  Enum := nil;
  objList := nil;
end;

// Dtruit l'instance ainsi que tous ses enfants.
destructor TWmiObject.Destroy;
var i : integer;
begin
  FWmiObject := nil;
  for i := 0 to FProperties.Count -1 do
  begin
    FProperties.Objects[i].free;
  end;
  FProperties.Free;

  for i := 0 to FChilds.Count -1 do
  begin
    TObject(FChilds[i]).free;
  end;
  FChilds.Free;

  inherited Destroy;
end;

function TWmiObject.GetClassName2: string;
begin
  if (FClassName2='') and (Assigned(FWmiObject))
  then FClassName2 := FWmiObject.Path_.Class_;
  result := FClassName2;
end;

function TWmiObject.GetDescription: string;
begin
  if (FDescription = '') and (Assigned(FWmiObject))
  then begin
    try
      FDescription := FWmiObject.Qualifiers_.Item('DisplayName', 0).Get_Value;
    except
      try
        FDescription := FWmiObject.Qualifiers_.Item('Description', 0).Get_Value;
      except
        FDescription := ClassName2;
      end;
    end;
  end;
  result := FDescription;
end;

// Retourne le nom associ au type d'vnement <EventType>. Si aucun nom
// n'a t dfini, GetEventTypeName retourne EventType sous la forme d'une
// chane de caractres.
function TWmiObject.GetEventTypeName(EventType: integer): string;
var
  i : integer;
begin
  // Par dfaut, on retourne directement EventType.
  result := IntToStr(EventType);

  // Il ne reste plus qu' chercher si un nom a t dfinit pour EventType.
  for i := 0 to high(EventTypes) do
  begin
    if EventTypes[i] = EventType
    then begin // EventType a t trouv dans EventTypes, on regarde s'il existe
               // un nom au mme index dans EventTypeNames.
      if i<=high(EventTypeNames)
      then result := EventTypeNames[i];
      break; // Dans tous les cas, la recherche est termine.
    end;
  end;
end;

// Dcode les donnes utilisateur d'un vnement ETW et les met en forme
// dans une chane de caractres.
// <MofData> doit pointer sur les donnes utilisateur de l'vnement.
// <MofLength> doit indiquer la taille total des donnes pointes par
// <MofData>.
function TWmiObject.GetUser(MofData : pointer; MofLength : integer) : string;
var
    i : integer;
    offset : cardinal;
    Field : TMofField;
    s : string;
begin
  if assigned(MofData) and (MofLength>0)
  then begin
    // Il faut parcourir toutes les proprits dans l'ordre de leur WmiDataId 
    result := '';
    if Properties.Count>0
    then begin
      Offset := 0;
      for i := 0 to Properties.Count -1 do
      begin
        // On demande  chaque champ de se dcoder lui mme.
        Field := TMofField(Properties.Objects[i]);
        s := Field.GetValue(MofData, MofLength, offset);

        // Si le champ est marqu avec NoPrint, on ne l'ajoute pas au rsultat.
        if (s<>'') and (not sameText(Field.Extension, 'NoPrint'))
        then result := result + s + ', ';
      end;

      // On supprime le dernier ', '
      if Result<>''
      then SetLength(Result, length(Result)-2);
    end;
  end
  else result := '';
end;

{ TMofField }

function TMofField.ArrayResult(Valeur: string): string;
begin
  if IsArray
  then result := '[' + Valeur + ']'
  else result := valeur;
end;

// Cre et initialise une nouvelle instance de la claase TMofField
constructor TMofField.create(prop: ISWbemProperty);
var
  Enum : IEnumVARIANT;
  TempObj : OleVariant;
  Fetch : Cardinal;
  Qualifier : ISWbemQualifier;
  v : OleVariant;
  low, high : integer;
  i : integer;

begin
  // On commence par initialiser les valeurs par dfaut.
  FDataType := prop.CIMType;
  FIsArray := prop.IsArray;
  FPropertyName := prop.Name;
  FDisplayLabel := FPropertyName;
  FIsPointer := false;
  FArraySize := 1;
  FExtension := '';
  FFormat := '';
  FStringTermination := '';
  FXMLFragment := false;
  FValueType := 'index';
  FIsPointer := false;
  
  // Puis on les remplace ventuellement en fonction des qualifiers trouvs.
  Enum := (prop.Qualifiers_._NewEnum) as IEnumVARIANT;
  while Enum.Next(1, TempObj, Fetch) = S_OK do
  begin
    Qualifier := IUnknown(TempObj) as ISWbemQualifier;
    if SameText(Qualifier.Name, 'WmiDataId')
    then FWmiDataId := Qualifier.Get_Value;

    if SameText(Qualifier.Name, 'Description')
    then FDisplayLabel := Qualifier.Get_Value;

    if SameText(Qualifier.Name, 'Pointer')
    then FIsPointer := true;

    if SameText(Qualifier.Name, 'MAX')
    then ArraySize := Qualifier.Get_Value;

    if SameText(Qualifier.Name, 'Extension')
    then FExtension := Qualifier.Get_Value;

    if SameText(Qualifier.Name, 'Format')
    then FFormat := Qualifier.Get_Value;

    if SameText(Qualifier.Name, 'StringTermination')
    then FStringTermination := Qualifier.Get_Value;

    if SameText(Qualifier.Name, 'XMLFragment')
    then FXMLFragment := true;

    if SameText(Qualifier.Name, 'ValueType')
    then FValueType := Qualifier.Get_Value;

    if SameText(Qualifier.Name, 'Pointer')
    then FIsPointer := true;


    if SameText(Qualifier.Name, 'ValueMap')
    then begin
      v := Qualifier.Get_Value;
      low := VarArrayLowBound(v, 1);
      high := VarArrayHighBound(v, 1);
      SetLength(ValueMap, high-low+1);
      for i := low to high do
      begin
        ValueMap[i-low] := VarArrayGet(v, [i]);
      end;
    end;

    if SameText(Qualifier.Name, 'Values')
    then begin
      v := Qualifier.Get_Value;
      low := VarArrayLowBound(v, 1);
      high := VarArrayHighBound(v, 1);
      SetLength(Values, high-low+1);
      for i := low to high do
      begin
        Values[i-low] := VarArrayGet(v, [i]);
      end;
    end;    

    if SameText(Qualifier.Name, 'BitMap')
    then begin
      v := Qualifier.Get_Value;
      low := VarArrayLowBound(v, 1);
      high := VarArrayHighBound(v, 1);
      SetLength(BitMap, high-low+1);
      for i := low to high do
      begin
        BitMap[i-low] := VarArrayGet(v, [i]);
      end;
    end;

    if SameText(Qualifier.Name, 'BitValues')
    then begin
      v := Qualifier.Get_Value;
      low := VarArrayLowBound(v, 1);
      high := VarArrayHighBound(v, 1);
      SetLength(BitValues, high-low+1);
      for i := low to high do
      begin
        BitValues[i-low] := VarArrayGet(v, [i]);
      end;
    end;
  end;
  if (DataType = wbemCimtypeObject) and (SameText(Extension, 'RWString'))
  then begin
    DataType := wbemCimtypeString;
    FormatQualifer := 'w';
    StringTermination := 'NoCounted';
  end;

  if (DataType = wbemCimtypeObject) and (SameText(Extension, 'RString'))
  then begin
    DataType := wbemCimtypeString;
    FormatQualifer := '';
    StringTermination := 'NoCounted';
  end;
end;

// Formate la valeur passe en paramtre en fonction des options de
// formatage dfinies pour la proprit. En particulier, remplace les
// valeurs de flag par les constantes associes, formate les types
// numrs...
function TMofField.FormatVal(valeur: cardinal): string;
var i : integer;
begin
  if SameText(FormatQualifer, 'c') // Afficher comme un caractre
  then result := chr(valeur)
  else begin
    if SameText(Extension, 'IPAddr') // Afficher comme une adresse IP
    then result := Format('%d.%d.%d.%d', [valeur and $FF,
                                          (valeur shr 8) and $FF,
                                          (valeur shr 16) and $FF,
                                          (valeur shr 24) and $FF])
    else begin
      if SameText(Extension, 'NoPrint ')
      then result := '' // Ne pas afficher.
      else begin
        if SameText(FormatQualifer, 'x') // Afficher en Hexadecimal.
        then result := Format('0x%0.8x', [valeur])
        else begin
          if high(BitMap) >=0 // Masque de bits dfinis.
          then begin
            result := '[';
            for i := 0 to high(BitMap) do
            begin
              if ((valeur shr BitMap[i]) and 1)<>0
              then result := result + BitValues[i] + ', ';
            end;
            if result <>'['
            then Setlength(result, length(result)-2);
            result := result + ']';
          end
          else begin
            if high(ValueMap)>=0  // Liste de flags
            then begin
              if FValueType = 'flag'
              then begin
                result := '';
                for i := 0 to high(ValueMap) do
                begin
                  if (Valeur and ValueMap[i])<>0
                  then result := result + Values[i] + ' or ';
                end;
                if result<>''
                then SetLength(result, length(result)-2);
              end
              else begin
                result := '';
                for i := 0 to high(ValueMap) do
                begin
                  if Valeur = ValueMap[i]
                  then result := result + Values[i] + ' or ';
                end;
                if result<>''
                then SetLength(result, length(result)-2);
              end;
            end
            else begin
              // Affichage brut. 
              result := IntToStr(valeur);
            end;
          end;
        end;
      end;
    end;
  end;
end;

// Extrait et formate la valeur du champ depuis le buffer de l'vnement
// <MofData>.
// <MofLength> indique la taille total du buffer point par <MofData>.
// En entre, <Offset> doit dsigner la position du champ par rapport a
// <MofData>. En sortie, <Offset> contient la position du prochaine champ
// dans <MofData>.
function TMofField.GetValue(MofData : pointer; MofLength : cardinal;
  var Offset : cardinal): string;
var
    ptr : pointer;
begin
  try
    result := '';

    // Calcul de l'adresse du champ.
    cardinal(ptr) := cardinal(MofData) + Offset;

    if IsPointer
    then begin
      // Si le champ est un pointeur, on ignore les autres information de type
      // qui s'applique en fait  la valeur pointe. On ne cherche pas 
      // interprter la valeur pointe, on affiche simplement la valeur du
      // pointeur en hexadcimal.
      result := Format('%0.4x', [cardinal(ptr^)]);
      inc(Offset, 4); // On avance de la taille du pointeur.
    end
    else begin
      if SameText(FExtension, 'Guid') and  // Les guid sont dclars sous
         (FDataType = wbemCimtypeUint8) and// la forme d'un tableau de 16 octets
         (ArraySize = 16)
      then begin
        result := GuidToString(TGUID(ptr^));
        inc(Offset, 16); // On avance de la taille du Guid.
      end
      else begin
        // Pour tous les autres type, on dlgue le dcodage du champ  une
        // mthode spcialise par type.
        case FDataType of
          wbemCimtypeSint32:
            Result := FormatSInt32(Offset, MofData);
          wbemCimtypeUint32:
            result := FormatUInt32(Offset, MofData);
          wbemCimtypeSint64, wbemCimtypeUint64:
            result := FormatInt64(Offset, MofData);
          wbemCimtypeString:
            result := FormatString(Offset, MofData, MofLength);
          wbemCimtypeBoolean:
            result := FormatBoolean(Offset, MofData);
          wbemCimtypeSint8:
            result := FormatSint8(Offset, MofData);
          wbemCimtypeUint8:
            result := FormatUInt8(Offset, MofData);
          wbemCimtypeSint16:
            result := FormatSInt16(Offset, MofData);
          wbemCimtypeUint16:
            result := FormatUInt16(Offset, MofData);
          wbemCimtypeObject:
            result := FormatObject(Offset, MofData);
          else raise Exception.Create('Type CIM non gr : ' + IntToStr(FDataType));
        end;
      end;
    end;
  except
    on e:exception do
      // En cas d'erreur, on renvoie le message d'erreur  la place de la valeur
      result := e.message;
  end;
end;

// Formate une donne de type word.
function TMofField.FormatUInt16(var Offset: Cardinal; MofData: Pointer) : string;
var
  i: Integer;
  valeur: Cardinal;
  ptr : pointer;
begin
  result := '';

  // Si le champ est un tableau, il faut traiter tous les lments
  // individuellement.
  for i := 0 to ArraySize - 1 do
  begin
    // Calcul de l'adresse du champ.  
    cardinal(ptr) := cardinal(MofData) + Offset;
    valeur := word(ptr^);
    result := result + FormatVal(valeur) + ', ';
    inc(Offset, 2); // On avance de la taille du champ.
  end;
  // Il faut enlever le dernier ', ' ajout au rsultat.  
  SetLength(result, length(result) - 2);

  result := ArrayResult(result);
end;

// Formate une donne de type word.
function TMofField.FormatSInt16(var Offset: Cardinal; MofData: Pointer) : string;
var
  i: Integer;
  ptr: pointer;
begin
  result := '';

  // Si le champ est un tableau, il faut traiter tous les lments
  // individuellement.
  for i := 0 to ArraySize - 1 do
  begin
    // Calcul de l'adresse du champ.
    cardinal(ptr) := cardinal(MofData) + Offset;
    result := result + IntToStr(smallint(ptr^)) + ', ';
    inc(Offset, 2); // On avance de la taille du champ.
  end;

  // Il faut enlever le dernier ', ' ajout au rsultat.  
  SetLength(result, length(result) - 2);

  result := ArrayResult(result);
end;

// Formate une donne de type byte.
function TMofField.FormatUInt8(var Offset: Cardinal; MofData: Pointer) : string;
var
  i: Integer;
  valeur: Cardinal;
  ptr : pointer;
begin
  result := '';

  // Si le champ est un tableau, il faut traiter tous les lments
  // individuellement.
  for i := 0 to ArraySize - 1 do
  begin
    // Calcul de l'adresse du champ.
    cardinal(ptr) := cardinal(MofData) + Offset;
    valeur := byte(ptr^);
    result := result + FormatVal(valeur) + ', ';
    inc(Offset, 1); // On avance de la taille du champ.
  end;

  // Il faut enlever le dernier ', ' ajout au rsultat.  
  SetLength(result, length(result) - 2);

  result := ArrayResult(result);
end;

// Formate une donne de type shortint.
function TMofField.FormatSint8(var Offset: Cardinal; MofData: Pointer) : string;
var
  i: Integer;
  ptr : pointer;
begin
  result := '';

  // Si le champ est un tableau, il faut traiter tous les lments
  // individuellement.
  for i := 0 to ArraySize - 1 do
  begin
    // Calcul de l'adresse du champ.
    cardinal(ptr) := cardinal(MofData) + Offset;
    result := result + IntToStr(shortint(ptr^)) + ', ';
    inc(Offset, 1); // On avance de la taille du champ.
  end;

  // Il faut enlever le dernier ', ' ajout au rsultat.
  SetLength(result, length(result) - 2);

  result := ArrayResult(result);
end;

// Formate une donne de type boolean
function TMofField.FormatBoolean(var Offset: Cardinal; MofData: Pointer) : string;
var
  i: Integer;
  ptr : pointer;
begin
  result := '';

  // Si le champ est un tableau, il faut traiter tous les lments
  // individuellement.
  for i := 0 to ArraySize - 1 do
  begin
    // Calcul de l'adresse du champ.
    cardinal(ptr) := cardinal(MofData) + Offset;
    result := result + BoolToStr(BOOL(ptr^), true) + ', ';

    // On avance de la taille du boolean.
    inc(offset, sizeof(bool));
  end;

  // Il faut enlever le dernier ', ' ajout au rsultat.
  SetLength(result, length(result) - 2);

  result := ArrayResult(result);
end;


// Formate une donne de type string.
function TMofField.FormatString(var Offset: Cardinal; MofData: Pointer;
  MofLength: Cardinal) : string;
var
  isWideString: Boolean;
  isNullTerminated: Boolean;
  strLength: Word;
  i: Integer;
  ptr : pointer;
  valStr: string;
  valWide: WideString;
begin
  // On regarde s'il s'agit d'un champ unicode. Les champs unicode ont un
  // indicateur FormatQualifier  w. Dans tous les autres cas, FormatQualifier
  // n'est pas dfinit.
  isWideString := FormatQualifer <> '';

  // Par dfaut, on considre qu'il s'agit d'une chane avec 0 terminal.
  isNullTerminated := true;

  // Calcul de l'adresse du champ.
  cardinal(ptr) := cardinal(MofData) + Offset;

  // Si la chane n'est pas  0 terminal, il faut d'abord dcoder sa longueur.
  strLength := 0;
  if SameText(StringTermination, 'Counted') or SameText(StringTermination, 'ReverseCounted')
  then begin
    // Le champ commence par un word qui indique la longueur de la chane.
    strLength := word(ptr^);
    inc(Offset, 2); // On saute la longueur de la chane.

    // Si la longueur est dfinie avec ReverseCounted, il faut inverser l'octet
    // de poid faible et l'octet de poid fort de la longueur lue.
    if SameText(StringTermination, 'ReverseCounted')
    then strLength := swap(strLength);
  end;

  if SameText(StringTermination, 'NotCounted')
  then begin
    // Aucune longueur de chane indique. Dans ce cas, la chane va jusqu' la
    // fin du buffer.
    isNullTerminated := false;
    strLength := MofLength - Offset;
  end;

  result := '';
  // Si le champ est un tableau, il faut traiter tous les lments
  // individuellement.
  for i := 0 to ArraySize - 1 do
  begin
    // Calcul de l'adresse du champ.
    cardinal(ptr) := cardinal(MofData) + Offset;
    if isWideString
    then begin // Cas des chane unicode.
      if isNullTerminated
      then begin
        valWide := PWideChar(ptr);
        result := result + QuotedStr(valWide) + ', ';
        // On avance de la longueur de la chane plus le 0 terminal (2 octets).
        inc(Offset, length(valWide) * 2 + 2);
      end
      else begin
        // On dfinit une chane de la longueur spcifie
        SetLength(valWide, strLength);

        // Et on recopie les donnes.
        move(ptr^, valWide[1], strLength * 2);
        result := result + QuotedStr(valWide) + ', ';

        // Cette fois, il ne faut avancer que le la longueur de la chane.
        // Comme on est en WideChar, chaque caractre fait 2 octets.
        inc(Offset, strLength * 2);
      end;
    end
    else begin
      // Cas des chanes ANSI.
      if isNullTerminated
      then begin
        valStr := PChar(ptr);
        result := result + QuotedStr(valStr) + ', ';

        // On avance de la longueur de la chane plus le 0 terminal.
        inc(Offset, length(valStr) + 1);
      end
      else begin
        // On dfinit une chane de la longueur spcifie
        SetLength(valStr, strLength);

        // Et on recopie les donnes.
        move(ptr^, valStr[1], strLength);
        result := result + QuotedStr(valStr) + ', ';

        // On avance de la longueur de la chane.
        inc(Offset, strLength);
      end;
    end;
  end;

  // Il faut enlever le dernier ', ' ajout au rsultat.
  SetLength(result, length(result) - 2);

  result := ArrayResult(result);
end;

// Formate une donne de type int64.
function TMofField.FormatInt64(var Offset: Cardinal; MofData: Pointer) : string;
var
  val64: Int64;
  i: Integer;
  ptr : pointer;
begin
  result := '';

  // Si le champ est un tableau, il faut traiter tous les lments
  // individuellement.
  for i := 0 to ArraySize - 1 do
  begin
    // Calcul de l'adresse du champ.
    cardinal(ptr) := cardinal(MofData) + Offset;
    inc(Offset, 8); // On avance de 8 octets dans le buffer MofData.
    val64 := int64(ptr^);

    // La valeur est crite en hexadcimal.
    result := result + Format('0x%0.16x, ', [val64]);
  end;
  
  // Il faut enlever le dernier ', ' ajout au rsultat.  
  SetLength(result, length(result) - 2);

  result := ArrayResult(result);
end;

// Formate une donne de type Objet.
function TMofField.FormatObject(var Offset: Cardinal; MofData: Pointer) : string;
var
  ip : cardinal;
  ptr : pointer;
begin
  result := '';
  // Calcul de l'adresse du champ.
  cardinal(ptr) := cardinal(MofData) + Offset;

  if SameText(FExtension, 'Guid') // L'objet est un Guid. On l'affiche tel quel.
  then begin
    result := GuidToString(TGUID(ptr^));
    Offset := Offset + 16; // On avance de la taille du champ.
  end;

  if SameText(FExtension, 'IPAddr') // L'objet est une adresse IP.
  then begin
    ip := cardinal(ptr^);
    inc(Offset, 4);  // On avance de la taille du champ.
    result := Format('%d.%d.%d.%d', [ip and $FF,
                                     (ip shr 8) and $FF,
                                     (ip shr 16) and $FF,
                                     (ip shr 24) and $FF]);
  end;
  if SameText(FExtension, 'Port')
  then begin // L'objet est un N de port TCP/IP
    result := IntToStr(word(ptr^));
    inc(Offset, 2); // On avance de la taille du champ.
  end;
  if SameText(FExtension, 'Sid')
  then begin // L'objet est un Sid
    result := 'Sid'; // TODO : Dcoder rellement la valeur du Sid.
    // On avance de la taille du champ
    inc(Offset, 4);
    inc(Offset, SECURITY_MAX_SID_SIZE);
  end;
end;

// Formate une donne de type Cardinal.
function TMofField.FormatUInt32(var Offset: Cardinal; MofData: pointer) : string;
var
  i: Integer;
  ptr : pointer;
begin
  result := '';

  // Si le champ est un tableau, il faut traiter tous les lments
  // individuellement.
  for i := 0 to ArraySize - 1 do
  begin
    // Calcul de l'adresse du champ.
    cardinal(ptr) := cardinal(MofData) + Offset;

    result := result + FormatVal(cardinal(ptr^)) + ', ';
    inc(Offset, 4); // On avance de 4 octets dans le buffer MofData.
  end;

  // Il faut enlever le dernier ', ' ajout au rsultat.
  SetLength(result, length(result) - 2);

  result := ArrayResult(result);
end;

// Formate une donne de type Integer.
function TMofField.FormatSInt32(var Offset: Cardinal; MofData: Pointer) : string;
var
  i: Integer;
  ptr : pointer;
begin
  result := '';

  // Si le champ est un tableau, il faut traiter tous les lments
  // individuellement.
  for i := 0 to ArraySize - 1 do
  begin
    // Calcul de l'adresse du champ.
    cardinal(ptr) := cardinal(MofData) + Offset;

    // On retourne simplement la valeur en chane.
    result := result + IntToStr(integer(ptr^)) + ', ';
    inc(Offset, 4); // On avance de 4 octets dans le buffer MofData.
  end;

  // Il faut enlever le dernier ', ' ajout au rsultat.
  SetLength(result, length(result) - 2);

  result := ArrayResult(result);
end;

// FormatHex est une fonction utilitaire qui permet de dumper le buffer <MofData>
// sous la forme d'une chane de caractres hexadcimale.
// <MofLength> doit dsigner la taille du buffer point par <MofData>.
function FormatHex(MofData : pointer; MofLength : integer) : string;
var i : integer;
begin
  result := '';
  for i := 0 to MofLength-1 do
  begin
    result := result + Format('%0.2x', [byte(MofData^)]);
    integer(MofData) := integer(MofData) + 1;
  end;
end;

initialization
finalization
  UniqueInstance.Free;
end.
