{******************************************************************************}
{                                                                              }
{ Event Trace information unit                                                 }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License          }
{ Version 1.1 (the "License"); you may not use this file except in             }
{ compliance with the License. You may obtain a copy of the License at         }
{ http://www.mozilla.org/MPL/                                                  }
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS"          }
{ basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the      }
{ License for the specific language governing rights and limitations           }
{ under the License.                                                           }
{                                                                              }
{ The Original Code is uEventTrace.pas released February 2008.                 }
{                                                                              }
{ The Initial Developer of the Original Code is Franck SORIANO.                }
{                                                                              }
{ Alternatively, the contents of this file may be used under the terms of the  }
{ GNU Lesser General Public License (the  "LGPL License"), in which case the   }
{ provisions of the LGPL License are applicable instead of those above.        }
{ If you wish to allow use of your version of this file only under the terms   }
{ of the LGPL License and not to allow others to use your version of this file }
{ under the MPL, indicate your decision by deleting  the provisions above and  }
{ replace  them with the notice and other provisions required by the LGPL      }
{ License.  If you do not delete the provisions above, a recipient may use     }
{ your version of this file under either the MPL or the LGPL License.          }
{                                                                              }
{ For more information about the LGPL: http://www.gnu.org/copyleft/lesser.html }
{                                                                              }
{******************************************************************************}
unit uEventTrace;

interface
uses classes, sysutils,
     uETW;

type
  // Reprsente un vnement de type EventTrace dans ETW
  // Il s'agit de la classe de base pour tous les vnements ETW. Elle modlise
  // les informations de l'entete de chaque vnement.
  TEventTrace = class(TObject)
  private
    FPID: cardinal;
    FThreadId: cardinal;
    FEventType: byte;
    FTimeStamp: int64;
    FEventVersion: word;
    FEventLevel: byte;
    FEventClass: TGUID;
  public
    // Cre et initialise une nouvelle instance de la classe TEventTrace.
    // <pEvent> est l'vnement de la trace ETW tel qu'il est fournit  la
    // mthode callback lors du traitement de ProcessTrace.
    constructor Create(pEvent : PEVENT_TRACE); virtual;

    // Renvoie le nom de l'vnement sous forme literale
    function GetEventClass : string; virtual;

    // Renvoie le type de l'vnement sous forme literale
    function GetEventType : string; virtual;

    // Renvoie le niveau de trace sous forme literale
    function GetLevel : string; virtual;

    // Retourne une chane de caractres reprsentant uniquement les donnes
    // utilisateur.
    function MofFieldToString : string; virtual;

    // Identifiant du processus ayant gnr l'vnement.
    property PID : cardinal read FPID;

    // Identifiant du Thread ayant gnr l'vnement.
    property ThreadId : cardinal read FThreadId;

    // Timestamp au moment de la gnration de l'vnement.
    property TimeStamp : int64 read FTimeStamp;

    // Guid de la classe identifiant l'vnement.
    property EventClass : TGUID read FEventClass;

    // Type de l'vnement.
    property EventType : byte read FEventType;

    // Niveau de trace de l'vnement.
    property EventLevel : byte read FEventLevel;

    // Numro de version de l'vnement.
    property EventVersion : word read FEventVersion;

  end;

  // TSQLEvent reprsente un vnement gnr par le logger SQL.
  TDefaultEvent = class(TEventTrace)
  private
    FOperationTime: int64;
    FMessage: string;
  public
    // Cre et initialise une nouvelle instance de la classe TSQLEvent.
    // <pEvent> est l'vnement de la trace ETW tel qu'il est fournit  la
    // mthode callback lors du traitement de ProcessTrace.
    constructor Create(pEvent : PEVENT_TRACE); override;

    // Renvoie le nom de l'vnement sous forme literale
    function GetEventClass : string; override;

    // Renvoie le type de l'vnement sous forme literale
    function GetEventType : string; override;

    // Retourne une chane de caractres reprsentant uniquement les donnes
    // utilisateur.
    function MofFieldToString : string; override;

    // Dure de l'opration.
    property OperationTime : int64 read FOperationTime;

    // Message crit dans la trace.
    property Msg : string read FMessage;
  end;

  TSQLEvent = class(TDefaultEvent)
  public
    // Renvoie le nom de l'vnement sous forme literale
    function GetEventClass : string; override;

    // Renvoie le type de l'vnement sous forme literale
    function GetEventType : string; override;

  end;

implementation
const
   LevelStr : array[0..5] of string=('0 - TRACE_LEVEL_NONE',
                                     '1 - TRACE_LEVEL_FATAL',
                                     '2 - TRACE_LEVEL_ERROR',
                                     '3 - TRACE_LEVEL_WARNING',
                                     '4 - TRACE_LEVEL_INFORMATION',
                                     '5 - TRACE_LEVEL_VERBOSE');

{ TEventTrace }

// Cre et initialise une nouvelle instance de la classe TTraceData.
// <pEvent> est l'vnement de la trace ETW tel qu'il est fournit  la
// mthode callback lors du traitement de ProcessTrace.
constructor TEventTrace.Create(pEvent: PEVENT_TRACE);
begin
  FPID := pEvent^.Header.ProcessId;
  FTimeStamp := pEvent^.Header.TimeStamp;
  FEventClass := pEvent^.Header.Guid;
  FEventVersion := pEvent^.Header.Version;
  FEventType := pEvent^.Header.Type_;
  FEventLevel := pEvent^.Header.Level;
  FThreadId := pEvent^.Header.ThreadId;
end;

// Renvoie le nom de l'vnement sous forme literale
function TEventTrace.GetEventClass: string;
begin
  // Par dfaut, on renvoie le GUID. Les classes drives pourront surcharger
  // la mthode pour retourner le nom rel de l'vnement.
  result := GUIDToString(FEventClass);
end;

// Renvoie le type de l'vnement sous forme literale
function TEventTrace.GetEventType: string;
begin
  // Par dfaut, on renvoie EventType. Les classes drives pourront surcharger
  // la mthode pour retourner le libell du type de l'vnement.
  result := IntToStr(FEventType);
end;

// Renvoie le niveau de trace sous forme literale
function TEventTrace.GetLevel: string;
begin
  if FEventLevel >high(LevelStr)
  then result := ''
  else result := LevelStr[FEventLevel];
end;

// Retourne une chane de caractres reprsentant uniquement les donnes
// utilisateur.
function TEventTrace.MofFieldToString: string;
begin
  result := ''; // Pas de donnes utilisateur pour EventTrace.
end;

// Renvoie le nom de l'vnement sous forme literale
function TDefaultEvent.GetEventClass: string;
begin
  result := 'EVENT';
end;

// Renvoie le type de l'vnement sous forme literale
function TDefaultEvent.GetEventType: string;
const
  DefaultEventType : array[0..3] of string = ('INFO', 'START', 'END', 'ERROR');
begin
  if EventType<=high(DefaultEventType)
  then result := DefaultEventType[EventType]
  else result := IntToStr(EventType);
end;

// Retourne une chane de caractres reprsentant uniquement les donnes
// utilisateur.
function TDefaultEvent.MofFieldToString: string;
begin
  result := Msg; // On ne renvoie que Msg.
end;

// Cre et initialise une nouvelle instance de la classe TDefaultEvent.
// <pEvent> est l'vnement de la trace ETW tel qu'il est fournit  la
// mthode callback lors du traitement de ProcessTrace.
constructor TDefaultEvent.Create(pEvent: PEVENT_TRACE);
type
  TSQLRec = packed record
    OperationTime : int64;
    msg : array[0..0] of char;
  end;
begin
  // On initialise les proprits de l'entete en appelant le constructeur
  // hrit.
  inherited Create(pEvent);

  // Enfin on dcode les informations dans MofData notre format propritaire
  FOperationTime := TSQLRec(pEvent^.MofData^).OperationTime; // Dure
  FMessage := PChar(@TSQLRec(pEvent^.MofData^).msg[0]); // Message
end;

// Renvoie le nom de l'vnement sous forme literale
function TSQLEvent.GetEventClass: string;
begin
  result := 'SQL Trace';
end;

// Renvoie le type de l'vnement sous forme literale
function TSQLEvent.GetEventType: string;
const
  SQLEventType : array[0..10] of string = ('INFO', 'START', 'END', 'ERROR',
     'BEGIN TRAN', 'ROLLBACK', 'COMMIT', 'PARAMETER START', 'PARAMETER END',
     'LOGIN', 'LOGOUT');
begin
  if EventType<=high(SQLEventType)
  then result := SQLEventType[EventType]
  else result := IntToStr(EventType);
end;

end.
