{******************************************************************************}
{                                                                              }
{ Event Tracing for Windows Logger Unit                                        }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License          }
{ Version 1.1 (the "License"); you may not use this file except in             }
{ compliance with the License. You may obtain a copy of the License at         }
{ http://www.mozilla.org/MPL/                                                  }
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS"          }
{ basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the      }
{ License for the specific language governing rights and limitations           }
{ under the License.                                                           }
{                                                                              }
{ The Original Code is uLogger.pas released February 2008.                     }
{                                                                              }
{ The Initial Developer of the Original Code is Franck SORIANO.                }
{                                                                              }
{ Alternatively, the contents of this file may be used under the terms of the  }
{ GNU Lesser General Public License (the  "LGPL License"), in which case the   }
{ provisions of the LGPL License are applicable instead of those above.        }
{ If you wish to allow use of your version of this file only under the terms   }
{ of the LGPL License and not to allow others to use your version of this file }
{ under the MPL, indicate your decision by deleting  the provisions above and  }
{ replace  them with the notice and other provisions required by the LGPL      }
{ License.  If you do not delete the provisions above, a recipient may use     }
{ your version of this file under either the MPL or the LGPL License.          }
{                                                                              }
{ For more information about the LGPL: http://www.gnu.org/copyleft/lesser.html }
{                                                                              }
{******************************************************************************}
unit uLogger;

interface
uses SysUtils, windows,
     uETW;

const
  // Identifiant du provider gnrique.
  DefaultLogger : TGUID = '{708A0B60-6786-460B-96A1-2C569C7B0B4C}';

  // Identifiant des classes de messages gnrs par le Logger gnrique.
  DefaultEventClass : TGUID = '{D3E38F07-FB65-4456-82B4-053A40FF72BD}';

const
  EVENT_INFO = 0;
  EVENT_START = 1;
  EVENT_END = 2;
  EVENT_ERROR = 3;

type
  TGenericLogger = class(TObject)
  private
    FEventEnabled : boolean;
    FEnableLevel : integer;
    FEnableFlags : cardinal;
    FTraceSession : TRACEHANDLE;
    FRegistrationHandle : TRACEHANDLE;
    FProviderGUID : TGUID;
    FEventClass : TGUID;
    FRegistrationEventClass : TRACE_GUID_REGISTRATION;
  protected
    // Indique si le message de niveau Level doit tre crit dans la trace en
    // cours. Cette fonction teste si une trace a t active ainsi que son
    // niveau d'activation. 
    function MessageEnabled(Level : byte) : boolean;

    // Active la gnration des vnements.
    procedure EnableEvents(Header : pointer);

    // Dsactive la gnration des vnements.
    procedure DisableEvents(Header : pointer);

    // Handle de la session de trace active.
    property TraceSession : TRACEHANDLE read FTraceSession;

    // Handle d'enregistrement du provider.
    property RegistrationHandle : TRACEHANDLE read FRegistrationHandle;

  public
    // Dclare le provider auprs de ETW.
    // <aProviderGUID> : GUID du provider.
    // <aEventClass> : GUID de la classe d'vnements gnr par le provider.
    constructor Create(aProviderGUID : TGUID; aEventClass : TGUID);

    // Dtruit la classe et annule l'enregistrement du provider auprs de ETW.
    destructor Destroy; override;

    // Ecrit le message <msg> dans la trace.
    procedure Trace(msg : string); overload;

    // Ecrit le message <msg> dans la trace.
    // Le message sera mmoris avec le type d'vnement <EventType>, pour le
    // niveau de trace <msgLevel>, avec une information de dure <OperationTime>
    procedure Trace(EventType : integer; msg : string;
                    msgLevel : integer = TRACE_LEVEL_INFORMATION;
                    OperationTime : int64=0); overload;

    // Ecrit le dbut d'une opration dans la trace :
    // Le message <msg> est crit dans la trace, avec le type d'vnement
    // <EventType> et le niveau de trace <msgLevel>.
    // En sortie, currentTime est initialise avec un TimeStamp correspondant 
    // l'heure courante au moment de l'appel.
    procedure TraceBegin(EventType : integer; msg : string;
                         var currentTime : int64;
                         msgLevel : integer = TRACE_LEVEL_INFORMATION);

    // Ecrit la fin d'une opration dans la trace :
    // Le message <msg> est crit dans la trace, avec le type d'vnement
    // <EventType> et le niveau de trace <msgLevel>.
    // TraceEnd doit avoir t prcd d'un appel  TraceBegin pour dfinir le
    // dbut de l'opration. <currentTime> doit contenir la valeur retourne par
    // le paramtre <currentTime> de TraceBegin.
    // TraceEnd calcule alors le temps coul (en ms) entre TraceEnd et
    // TraceBegin et crit ce temps pass dans le champ OperationTime de la
    // trace.
    // En sortie, currentTime contient le temps coul depuis l'appel 
    // TraceBegin 
    procedure TraceEnd(EventType : integer; msg : string; var currentTime : int64;
                       msgLevel : integer = TRACE_LEVEL_INFORMATION);

    // Ecrit l'exception <e> dans la trace, avec le niveau de trace <msgLevel>
    procedure TraceException(e : exception;
                             msgLevel : integer = TRACE_LEVEL_ERROR);

    // Identifiant du provider.
    property ProviderGUID : TGUID read FPRoviderGUID;

    // Classe de l'vnement gnr par le provider.
    property EventClass : TGUID read FEventClass;

    // Indique si la trace est actuellement active
    property EventEnabled : boolean read FEventEnabled write FEventEnabled;

    // Indique le niveau de trace qui a t dfinit au moment de l'activation
    // de la trace.
    property EnableLevel : integer read FEnableLevel;

    // Indique les flags dfinits au moment de l'activation de la trace
    property EnableFlags : cardinal read FEnableFlags;
  end;

// Retourne l'instance unique du logger gnrique.  
function GenericLogger : TGenericLogger;

implementation
type
  TGenericEvent = packed record
    Header : EVENT_TRACE_HEADER;        // Entte du message
    Data : array[0..1] of MOF_FIELD;    // Informations  crire dans la trace.
  end;

var Logger : TGenericLogger = nil; // Instance unique du logger gnrique.

// Retourne l'instance unique du logger gnrique.
function GenericLogger : TGenericLogger;
begin
  result := Logger;
end;

// Cette fonction est appele automatiquement par ETW pour controler la trace :
// <RequestCode> indique le type d'opration dsire :
//      - WMI_ENABLE_EVENTS : Nous informe que le provider vient d'tre activ.
//      - WMI_DISABLE_EVENTS : Nous informe que le provider vient d'tre arrt.
// <RequestContext> Contient un pointeur sur les donnes utilisateur fournies
//                  au moment de l'enregistrement de la trace avec
//                  RegisterTraceGuid.
//                  On l'utilise pour connaitre l'instance de TGenericLogger
//                  concerne.
// <BufferSize> Spcifique Vista. On l'ignore.
// <Buffer> pointe sur une structure WNODE_HEADER contenant des informations
//          sur la trace concerne.
// En retour cette fonction doit renvoyer ERROR_SUCCESS pour indiquer que tout
// s'est bien pass.
function TraceControlCallBack(RequestCode : WMIDPREQUESTCODE;
         RequestContext : pointer;
         BufferSize : cardinal;
         Buffer : pointer) : cardinal; stdcall;
begin
  // En fonction de l'opration demande, on active ou dsactive la trace.
  case RequestCode of
    WMI_ENABLE_EVENTS: TGenericLogger(RequestContext).EnableEvents(Buffer);
    WMI_DISABLE_EVENTS: TGenericLogger(RequestContext).DisableEvents(Buffer);
  end;
  result := ERROR_SUCCESS;
end;

// Ecrit le message <msg> dans la trace.
// Le message sera mmoris avec le type d'vnement <EventType>, pour le
// niveau de trace <msgLevel>, avec une information de dure <OperationTime>
procedure TGenericLogger.Trace(EventType: integer; msg: string;
                               msgLevel: integer;
                               OperationTime: int64);
const EmptyStr : char = #0;
var Event : TGenericEvent;
    EventSize : cardinal;
begin
  // Il faut commencer par tester si la trace a t active et si le niveau de
  // trace dfinit est compatible avec le message  tracer. Sinon on
  // n'enregistre pas le message dans la trace.
  if MessageEnabled(msgLevel)
  then begin
    // Initialisation d'une structure TGenericEvent qui sera crite dans la
    // trace.
    EventSize := Sizeof(Event);
    fillchar(Event, EventSize, 0);

    // Initialisation de l'entte de l'vnement.
    Event.Header.Size := EventSize; // Taille de la structure.

    // On indique que le champs Guid a t dfinit et qu'on utilise un tableau
    // de structure MOF_FIELD pour dfinir les informations crites dans la
    // trace.
    Event.Header.Flags := WNODE_FLAG_TRACED_GUID or WNODE_FLAG_USE_MOF_PTR;

    // On dfinit les identifiants de l'vnement qui sera gnr dans la trace.
    // Ces identfiants permettront au programme client de dcoder les
    // informations dfinies dans Data.
    Event.Header.Guid := FEventClass; // Dfinition de la classe d'vnement
    Event.Header.Version := 0;        // Numro de version de l'vnement
    Event.Header.Type_ := EventType; // Dfinition du type de l'vnement

    Event.Header.Level := msgLevel; // Niveau de trace de l'vnement.

    // Une fois l'entte dfinie, on dfinit les informations  crire dans la
    // trace.
    // On veut crire un enregistrement compos des champs <OperationTime,
    // msg> Pour cel, on initialise l'adresse et la longueur de chaque champ.
    Event.Data[0].DataPtr := @OperationTime;
    Event.Data[0].Length := sizeof(OperationTime);
    if msg<>''
    then begin
      Event.Data[1].DataPtr := @msg[1];
      Event.Data[1].Length := length(msg)+1; // Longueur de la chane plus zro
                                             // terminal.
    end
    else begin
      // Si msg est une chaine vide, msg est en ralit un pointeur qui vaut nil
      // Il ne s'agit donc plus d'une chaine asciiz et la chaine ne s'enregistre
      // pas correctement. Il faut donc qu'on crive explicitement un zro
      // terminal.
      Event.Data[1].DataPtr := @EmptyStr;
      Event.Data[1].Length := 1;
    end;

    // Enfin on crit le message dans la trace.
    TraceEvent(TraceSession, @Event);
  end;
end;

// Ecrit le message <msg> dans la trace.
procedure TGenericLogger.Trace(msg: string);
begin
  // On se contente d'appeler Trace en dfinissant un type de message gnrique.
  Trace(EVENT_TRACE_TYPE_INFO, msg);
end;

// Ecrit le dbut d'une opration dans la trace :
// Le message <msg> est crit dans la trace, avec le type d'vnement
// <EventType> et le niveau de trace <msgLevel>.
// En sortie, currentTime est initialise avec un TimeStamp correspondant 
// l'heure courante au moment de l'appel.
procedure TGenericLogger.TraceBegin(EventType: integer; msg: string;
  var currentTime: int64; msgLevel: integer);
begin
  // Tout d'abord, on crit le message normalement dans la trace.
  Trace(EventType, msg, msgLevel);

  // Ensuite on rcupre le timestamp dans currentTime
  QueryPerformanceCounter(currentTime);  
end;

// Ecrit la fin d'une opration dans la trace :
// Le message <msg> est crit dans la trace, avec le type d'vnement
// <EventType> et le niveau de trace <msgLevel>.
// TraceEnd doit avoir t prcd d'un appel  TraceBegin pour dfinir le
// dbut de l'opration. <currentTime> doit contenir la valeur retourne par
// le paramtre <currentTime> de TraceBegin.
// TraceEnd calcule alors le temps coul (en ms) entre TraceEnd et
// TraceBegin et crit ce temps pass dans le champ OperationTime de la
// trace.
// En sortie, currentTime contient le temps coul depuis l'appel 
// TraceBegin.
procedure TGenericLogger.TraceEnd(EventType: integer; msg: string;
  var currentTime: int64; msgLevel: integer);
var StartTime,
    EndTime,
    Freq,
    OperationTime : int64;

begin
  // On commence par lire le TimeStamp de fin.
  QueryPerformanceCounter(EndTime);
  QueryPerformanceFrequency(Freq);
  StartTime := currentTime;

  // On calcul le temps pass depuis TraceStart.
  OperationTime := (EndTime-StartTime) * 1000000 div freq;

  // Enfin on crit le message dans la trace, en indiquant le temps qu'on vient
  // de calculer.
  Trace(EventType, msg, msgLevel, OperationTime);
  currentTime := OperationTime;
end;

// Ecrit l'exception <e> dans la trace, avec le niveau de trace <msgLevel>
procedure TGenericLogger.TraceException(e: exception; msgLevel: integer);
begin
  // On se contente d'crire le message d'erreur dans la trace, avec
  // EVENT_TRACE_TYPE_EXCEPTION comme identifiant de message pour indiquer
  // qu'il s'agit d'une exception.
  Trace(EVENT_ERROR, e.ClassName + ': ' + e.Message, msgLevel);
end;

// Dsactive la gnration des vnements.
destructor TGenericLogger.Destroy;
begin
  UnregisterTraceGuids(FRegistrationHandle);
  inherited Destroy;
end;

procedure TGenericLogger.DisableEvents;
begin
  // La trace a t dsactive.
  EventEnabled := false;
  FTraceSession := 0;
  FEnableFlags := 0;
  FEnableLevel := 0;
end;

// Active la gnration des vnements.
procedure TGenericLogger.EnableEvents;
begin
  // On commence par rcuprer le Handle de la trace concerne
  FTraceSession := GetTraceLoggerHandle(Header);
  if TraceSession<>INVALID_HANDLE_VALUE // En cas d'erreur, on ignore la trace. 
  then begin
    // Lecture du niveau de trace active.
    FEnableLevel := GetTraceEnableLevel(TraceSession);

    // Lecture des flags dfinis avec la trace
    FEnableFlags := GetTraceEnableFlags(TraceSession);

    FEventEnabled := true; // La trace a t active.
  end;
end;

// Indique si le message de niveau Level doit tre crit dans la trace en
// cours. Cette fonction teste si une trace a t active ainsi que son
// niveau d'activation.
function TGenericLogger.MessageEnabled(Level: byte): boolean;
begin
  result := EventEnabled and ((Level<=EnableLevel) or (EnableLevel = 0));
end;

// Dclare le provider auprs de ETW.
// <aProviderGUID> : GUID du provider.
// <aEventClass> : GUID de la classe d'vnements gnr par le provider.
constructor TGenericLogger.Create(aProviderGUID: TGUID; aEventClass : TGUID);
begin
  FProviderGUID := aProviderGUID;
  FEventClass := aEventClass;
  FRegistrationEventClass.Guid := @FEventClass;
  FRegistrationEventClass.RegHandle := 0;

  // la dclaration du provider consiste  appeler RegisterTraceGuids.
  if RegisterTraceGuids(@TraceControlCallBack, // Mthode callback pour
                                               // contrler la trace
    Self, // On dfinit le pointeur sur l'instance comme contexte utilisateur.
    @FProviderGUID, // Pointeur sur l'identifiant du provider
    1, // Nombre d'vnements dclars
    @FRegistrationEventClass, // Pointeur sur la liste des classes d'vnements
                              // gnrs par le provider
    nil, // On ne dfinit pas d'informations MOF
    nil, // On ne dfinit pas d'informations MOF
    FRegistrationHandle)<>ERROR_SUCCESS
  then RaiseLastOSError;
end;

initialization
  // Le logger gnrique est initialis au dmarrage de l'application. Il
  // dfinit le Provider DefaultLogger qui gnre une seule classe d'vnement :
  // DefaultEventClass
  logger := TGenericLogger.Create(DefaultLogger, DefaultEventClass);
finalization
  // Dtruit le logger.
  logger.Free;
end.
