{******************************************************************************}
{                                                                              }
{ Event Tracing for Windows SQL Logger Unit                                    }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License          }
{ Version 1.1 (the "License"); you may not use this file except in             }
{ compliance with the License. You may obtain a copy of the License at         }
{ http://www.mozilla.org/MPL/                                                  }
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS"          }
{ basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the      }
{ License for the specific language governing rights and limitations           }
{ under the License.                                                           }
{                                                                              }
{ The Original Code is uSQLLogger.pas released February 2008.                  }
{                                                                              }
{ The Initial Developer of the Original Code is Franck SORIANO.                }
{                                                                              }
{ Alternatively, the contents of this file may be used under the terms of the  }
{ GNU Lesser General Public License (the  "LGPL License"), in which case the   }
{ provisions of the LGPL License are applicable instead of those above.        }
{ If you wish to allow use of your version of this file only under the terms   }
{ of the LGPL License and not to allow others to use your version of this file }
{ under the MPL, indicate your decision by deleting  the provisions above and  }
{ replace  them with the notice and other provisions required by the LGPL      }
{ License.  If you do not delete the provisions above, a recipient may use     }
{ your version of this file under either the MPL or the LGPL License.          }
{                                                                              }
{ For more information about the LGPL: http://www.gnu.org/copyleft/lesser.html }
{                                                                              }
{******************************************************************************}
unit uSQLLogger;

interface

uses sysutils, db,
     uLogger, uETW;

const
  // Identifiant du provider SQL.
  SQLLoggerGUID : TGUID = '{E356181E-4059-4BBA-BA4A-87280762A798}';

  // Identifiant des classes de messages gnrs par le provider SQL.
  SQLEventClass : TGUID = '{815AA09D-DBC9-45FE-982E-0243FDF363ED}';

// Constantes pour les types d'vnements  
const
  EVENT_SQL_INFO = 0;     // Message d'information
  EVENT_SQL_START = 1;    // Dbut d'excution d'une requte SQL
  EVENT_SQL_END = 2;      // Fin d'excution d'une requte SQL
  EVENT_SQL_ERROR = 3;    // Erreur survenue avec une requte SQL
  EVENT_BEGINTRAN = 4;    // Dbut d'une transaction.
  EVENT_ROLLBACKTRAN = 5; // Rollback d'une transaction.
  EVENT_COMMITTRAN = 6;   // Commit d'une transaction.
  EVENT_SQL_FORMAT_START = 7; // Dbut de la mise en forme des paramtres
  EVENT_SQL_FORMAT_END = 8; // Fin de la mise en forme des paramtres.
  EVENT_SQL_CONNECT = 9;     // Connexion au SGBD
  EVENT_SQL_DISCONNECT = 10; // Dexonnexion du SGBD

// TSQLLogger implmente un provider ETW pour gnrer des traces d'excution de
// requtes SQL.  
type
  TSQLLogger = class(TGenericLogger)
  private
    // Retourne la valeur du paramtre <Param> sous la forme d'une chane de
    // caractres.
    function ParamStr(Param : TParam) : string;

    // Ecrit la valeur des paramtres <Params> de la requte <SQL> dans la
    // trace. Cette criture s'effectue sous la forme de la requte <SQL> dans
    // laquelle les spcificateurs de paramtres ont t remplacs par la valeur
    // du paramtre correspondant.
    procedure FormatParams(SQL : string; params : TParams);
  public
    // Trace le dbut d'excution de la requte <SQL>.
    // En sortie, <Time> contient le timestamp du dbut d'excution de la
    // requte <Params> doit contenir la valeur des paramtres de la requte.
    // Ces derniers peuvent tre dfinits soit sous la forme d'un ? soit sous la
    // forme :NomdeParametre.
    // <msgLevel> doit indiquer le niveau de trace du message.
    // Si la trace est active au niveau VERBOSE, les valeurs des paramtres de
    // la requte seront crites dans la trace.
    procedure TraceSQLBegin(SQL : string; var Time : int64; Params : TParams=nil;
                            msgLevel : byte = TRACE_LEVEL_INFORMATION);

    // Trace la fin d'excution de la requte <SQL>.
    // <Time> doit contenir le timestamp du dbut d'excution de la requte
    //        renvoy par TraceSQLBegin.
    //        En sorti, <Time> contiendra le temps coul depuis l'appel 
    //        TraceSQLBegin.
    // <Params> doit contenir la valeur des paramtres de la requte.
    // Ces derniers peuvent tre dfinits soit sous la forme d'un ? soit sous la
    // forme :NomdeParametre.
    // <msgLevel> doit indiquer le niveau de trace du message.
    // Si la trace est active au niveau VERBOSE, les valeurs des paramtres de
    // la requte seront crites dans la trace.
    procedure TraceSQLEnd(SQL : string; var Time : int64; Params : TParams=nil;
                          msgLevel : byte = TRACE_LEVEL_INFORMATION);

    // Trace le dbut d'une transaction SQL.
    // En sortie, <Time> contient le timestamp du dbut de la transaction
    // <msgLevel> indique le niveau de trace du message.
    procedure TraceStartTransaction(var Time : int64;
                                    msgLevel : byte = TRACE_LEVEL_INFORMATION);

    // Trace la validation d'une transaction SQL.
    // <Time> doit contenir le timestamp du dbut de la transaction renvoy par
    //        TraceStartTransaction.
    // <msgLevel> indique le niveau de trace du message.
    procedure TraceCommitTransaction(Time : int64;
                                     msgLevel : byte = TRACE_LEVEL_INFORMATION);

    // Trace le rollback d'une transaction SQL.
    // <Time> doit contenir le timestamp du dbut de la transaction renvoy par
    //        TraceStartTransaction.
    // <msgLevel> indique le niveau de trace du message.
    procedure TraceRollbackTransaction(Time : int64;
                                     msgLevel : byte = TRACE_LEVEL_INFORMATION);

    // Trace la connexion au SGBD.
    procedure TraceConnect(msgLevel : byte = TRACE_LEVEL_INFORMATION);

    // Trace la dconnexion du SGBD.
    procedure TraceDisconnect(msgLevel : byte = TRACE_LEVEL_INFORMATION);
  end;

// Retourne l'instance unique du logger SQL.  
function SQLLogger : TSQLLogger;
implementation

var Logger : TSQLLogger = nil; // Instance unique du logger.
    InvariantCulture : TFormatSettings; // Paramtres pour le formattage des
                                        // paramtres SQL. Force le dcimal
                                        // sparateur  '.'

// Retourne l'instance unique du logger SQL.
function SQLLogger : TSQLLogger;
begin
  // Si l'instance unique n'existe pas, il faut la crer.
  if not Assigned(Logger)
  then Logger := TSQLLogger.Create(SQLLoggerGUID, SQLEventClass);

  result := Logger;
end;

{ TSQLLogger }

// Ecrit la valeur des paramtres <Params> de la requte <SQL> dans la
// trace. Cette criture s'effectue sous la forme de la requte <SQL> dans
// laquelle les spcificateurs de paramtres ont t remplacs par la valeur
// du paramtre correspondant.
procedure TSQLLogger.FormatParams(SQL: string; params: TParams);
var ParamNr : integer; 
    i : integer;
    Start : integer;
    result : string; // Requte formate avec la valeur des paramtres. 
    Time : int64;

  // Remplace le spcificateur de paramtre par la valeur du paramtre
  // correspondant dans la chaine de rsultat.
  procedure ReplaceParam;
  begin
    // On copie la portion de la requte qui se trouvait avant le paramtre.
    result := result + Copy(SQL, Start, i-Start); // La position i est exclue
                                                  // de la copie.

    // On crit la valeur du paramtre dans la chaine de rsultat.
    if ParamNr<Params.Count
    then result := result + ParamStr(Params[ParamNr])
    else result := result + '?';

    inc(ParamNr); // On passe au paramtre courant.
  end;

  // Saute le commentaire dans l'analyse de la requte.
  // <ch1><ch2> doivent indiquer les caractres attendu pour dfinir la fin du
  // commentaire.
  procedure SkipComment(ch1, ch2 : char);
  begin
    inc(i,2); // On saute le marqueur de dbut du commentaire.
    while (i<length(SQL)) and (SQL[i]<>ch1) and (SQL[i+1]<>ch2) do
    begin
      inc(i);
    end;
  end;

  // Saut le nom du paramtre dans l'analyse de la requte.
  procedure SkipParamName;
  begin
    while (i<=length(SQL)) and
            (
              ((SQL[i]>='0') and (SQL[i]<='9')) or
              ((SQL[i]>='A') and (SQL[i]<='Z')) or
              ((SQL[i]>='a') and (SQL[i]<='z')) or
               (SQL[i] = '_')
            ) do
    begin
      inc(i);
    end;
  end;


begin
  // On trace le dbut de la mise en forme, au niveau VERBOSE.
  TraceBegin(EVENT_SQL_FORMAT_START, '', Time, TRACE_LEVEL_VERBOSE);
  try
    try
      ParamNr := 0; // On commence avec le paramtre 0.
      Start := 1; // On se positionne sur le premier caractre.
      i := 1;
      result := '';

      // On parcour la requte SQL caractre par caractre pour reprer les
      // spcificateurs de paramtres.
      while i<=length(SQL) do
      begin
        case SQL[i] of
        '?' : // Emplacement d'un paramtre '?'
          begin
            ReplaceParam; // Remplace le ? par la valeur du paramtre.
            inc(i);
            Start := i;
          end;
        '/' : // Cas des commentaires /* */. Si c'est le cas, il faut sauter le
              // commentaire.
              if (i<length(SQL)) and (SQL[i+1] = '*')
              then SkipComment('*', '/')
              else inc(i);
        '-': // Cas des commentaires --. Si c'est le cas, il faut sauter le
              // commentaire.
             if (i<length(SQL)) and (SQL[i+1] = '-')
             then SkipComment(#13, #10)
             else inc(i);
        ':': // Cas d'un paramtre nomm
          begin
            ReplaceParam; // On remplace le paramtre par sa valeur
            inc(i);
            SkipParamName; // On saute le nom du paramtre.
            Start := i;
          end;
        else inc(i);
        end;
      end;
      if Start <= length(SQL)
      then result := result + Copy(SQL, Start, length(SQL));

    except
      // En cas d'erreur, on trace l'exception. Par contre, on ne redclenche
      // pas l'exception qui a t gnre  cause de l'criture dans la trace.
      on e:exception do
        TraceException(e);
    end;
  finally
    // Enfin on crit la requte formate dans la trace.
    TraceEnd(EVENT_SQL_FORMAT_END, result, Time, TRACE_LEVEL_VERBOSE);
  end;
end;

// Retourne la valeur du paramtre <Param> sous la forme d'une chane de
// caractres.
function TSQLLogger.ParamStr(Param: TParam): string;
var s : string;
    i : integer;
begin
  if Param.IsNull        // Si le paramtre est  null
  then result := 'null'  // on crit null comme valeur.
  else begin             // Sinon, il faut formater la valeur en fonction de son
                         // type.
    case Param.DataType of
    ftBoolean:
      // Pour les boolean, on crit 0 ou -1.
      result := BoolToStr(Param.AsBoolean);
    ftFixedWideChar, ftWideMemo, ftOraClob, ftFixedChar,
    ftWideString, ftMemo, ftString:
      // Pour les chanes de caractres, on crit la chaine entre quotes.
      result := QuotedStr(Param.AsString);
    ftLargeint, ftAutoInc, ftSmallint, ftInteger, ftWord :
      // Pour les entiers, on crit la valeur entire directement.
      result := IntToStr(Param.AsInteger);
    ftFMTBcd, ftFloat, ftCurrency, ftBCD:
      // Pour les nombres dcimaux, on crit un nombre avec un '.' comme
      // sparateur dcimal.
      result := FloatToStr(Param.AsFloat, InvariantCulture);
    ftDate:
      // Pour les dates, on formate la date entre quotes.
      result := QuotedStr(DateToStr(Param.AsDate, InvariantCulture));
    ftTime:
      // Pour les heures, on formate l'heure entre quotes.
      result := QuotedStr(TimeToStr(Param.AsTime, InvariantCulture));
    ftDateTime:
      // Pour les dates, on formate la date entre quotes.
      result := QuotedStr(DateTimeToStr(Param.AsDateTime, InvariantCulture));
    ftOraTimeStamp, ftOraInterval, ftTimeStamp, ftVariant, ftInterface,
    ftIDispatch, ftGuid, ftADT, ftArray, ftReference, ftDataSet, ftOraBlob,
    ftParadoxOle, ftDBaseOle, ftTypedBinary, ftCursor, ftGraphic, ftFmtMemo,
    ftBlob, ftUnknown, ftBytes, ftVarBytes:
      begin
        // Pour tous les autres types, on crit un dump hexadecimal
        s := Param.AsBlob;
        result := '0x';
        for i := 1 to length(s) do
        begin
          result := result + Format('%0.2x', [s[i]]);
        end;
      end;
    end;
  end;
end;

// Trace la validation d'une transaction SQL.
// <Time> doit contenir le timestamp du dbut de la transaction renvoy par
//        TraceStartTransaction.
// <msgLevel> indique le niveau de trace du message.
procedure TSQLLogger.TraceCommitTransaction(Time: int64; msgLevel: byte);
begin
  TraceEnd(EVENT_COMMITTRAN, '', Time, msgLevel);
end;

// Trace la connexion au SGBD.
procedure TSQLLogger.TraceConnect(msgLevel: byte);
begin
  Trace(EVENT_SQL_CONNECT, 'Login.', msgLevel);
end;

// Trace la dconnexion du SGBD.
procedure TSQLLogger.TraceDisconnect(msgLevel: byte);
begin
  Trace(EVENT_SQL_DISCONNECT, 'Logout.', msgLevel);
end;

// Trace la fin d'excution de la requte <SQL>.
// <Time> doit contenir le timestamp du dbut d'excution de la requte
//        renvoy par TraceSQLBegin.
//        En sorti, <Time> contiendra le temps coul depuis l'appel 
//        TraceSQLBegin.
// <Params> doit contenir la valeur des paramtres de la requte.
// Ces derniers peuvent tre dfinits soit sous la forme d'un ? soit sous la
// forme :NomdeParametre.
// <msgLevel> doit indiquer le niveau de trace du message.
// Si la trace est active au niveau VERBOSE, les valeurs des paramtres de
// la requte seront crites dans la trace.
procedure TSQLLogger.TraceSQLEnd(SQL: string; var Time: int64; Params: TParams;
  msgLevel : byte);
begin
  // On crit le message de fin de la requte brute.
  TraceEnd(EVENT_SQL_END, SQL, Time, msgLevel);

  // Si la trace est active pour le niveau VERBOSE, on crit galement la valeur
  // des paramtres.
  if MessageEnabled(TRACE_LEVEL_VERBOSE) and
        Assigned(Params) and (Params.Count >0)
  then FormatParams(SQL, Params);
end;

// Trace le rollback d'une transaction SQL.
// <Time> doit contenir le timestamp du dbut de la transaction renvoy par
//        TraceStartTransaction.
// <msgLevel> indique le niveau de trace du message.
procedure TSQLLogger.TraceRollbackTransaction(Time: int64; msgLevel: byte);
begin
  TraceEnd(EVENT_ROLLBACKTRAN, '', Time, msgLevel);
end;

// Trace le dbut d'excution de la requte <SQL>.
// En sortie, <Time> contient le timestamp du dbut d'excution de la
// requte <Params> doit contenir la valeur des paramtres de la requte.
// Ces derniers peuvent tre dfinits soit sous la forme d'un ? soit sous la
// forme :NomdeParametre.
// <msgLevel> doit indiquer le niveau de trace du message.
// Si la trace est active au niveau VERBOSE, les valeurs des paramtres de
// la requte seront crites dans la trace.
procedure TSQLLogger.TraceSQLBegin(SQL: string; var Time: int64;
  Params: TParams; msgLevel : byte);
begin
  // On commence par crire la requte brte.
  TraceBegin(EVENT_SQL_START, SQL, Time, msgLevel);

  // Si le niveau verbose est activ, on format les paramtre dans la requte.
  if MessageEnabled(TRACE_LEVEL_VERBOSE) and
        Assigned(Params) and (Params.Count >0)
  then FormatParams(SQL, Params);
end;

// Trace le dbut d'une transaction SQL.
// En sortie, <Time> contient le timestamp du dbut de la transaction
// <msgLevel> indique le niveau de trace du message.
procedure TSQLLogger.TraceStartTransaction(var Time: int64; msgLevel : byte);
begin
  TraceBegin(EVENT_BEGINTRAN, '', Time);
end;

initialization
  GetLocaleFormatSettings(0, InvariantCulture);
  InvariantCulture.DecimalSeparator := '.';
end.
