{******************************************************************************}
{                                                                              }
{ Event Trace information unit                                                 }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License          }
{ Version 1.1 (the "License"); you may not use this file except in             }
{ compliance with the License. You may obtain a copy of the License at         }
{ http://www.mozilla.org/MPL/                                                  }
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS"          }
{ basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the      }
{ License for the specific language governing rights and limitations           }
{ under the License.                                                           }
{                                                                              }
{ The Original Code is uTraceData.pas released February 2008.                  }
{                                                                              }
{ The Initial Developer of the Original Code is Franck SORIANO.                }
{                                                                              }
{ Alternatively, the contents of this file may be used under the terms of the  }
{ GNU Lesser General Public License (the  "LGPL License"), in which case the   }
{ provisions of the LGPL License are applicable instead of those above.        }
{ If you wish to allow use of your version of this file only under the terms   }
{ of the LGPL License and not to allow others to use your version of this file }
{ under the MPL, indicate your decision by deleting  the provisions above and  }
{ replace  them with the notice and other provisions required by the LGPL      }
{ License.  If you do not delete the provisions above, a recipient may use     }
{ your version of this file under either the MPL or the LGPL License.          }
{                                                                              }
{ For more information about the LGPL: http://www.gnu.org/copyleft/lesser.html }
{                                                                              }
{******************************************************************************}
unit uTraceData;

interface

uses windows, sysutils,
     uETW, uETWWmiInfo, uEventTrace;

type
  // Cette classe reprsente un vnement lu depuis une trace ETW
  TTraceData = class(TEventTrace)
  private
    FUserData : string;

    FThreadCol: cardinal;
    FEventName : string;
    FEventTypeName : string;
    FSetup : boolean;
  public
    // Contenu des donnes utilisateur de l'vnement
    UserData : array of byte;

    // Cre et initialise une nouvelle instance de la classe TTraceData.
    // <pEvent> est l'vnement de la trace ETW tel qu'il est fournit  la
    // mthode callback lors du traitement de ProcessTrace.
    constructor Create(pEvent : PEVENT_TRACE); override;

    // Fonction utilitaire qui dcode les donnes de UserData et les formate
    // pour qu'elles soient renvoye par les autres mthodes de la classe.
    procedure Setup;

    // Retourne les informations utilisateur, formatte sous la forme d'une
    // chane de caractres.
    function GetUserData : string;

    // Retourne l'instance TWmiObject qui dcrit la structure des donnes
    // utilisateur dans UserData. Si les informations MOF n'ont pas t publies
    // la mthode renvoie nil.
    function GetInfo : TWmiObject;

    // Renvoie le nom de l'vnement sous forme literale
    function GetEventClass : string; override;

    // Renvoie le type de l'vnement sous forme literale
    function GetEventType : string; override;


    // Rang du thread dans la trace.
    property ThreadCol : cardinal read FThreadCol write FThreadCol;
  end;
  
implementation

const
   LevelStr : array[0..5] of string=('0 - TRACE_LEVEL_NONE',
                                     '1 - TRACE_LEVEL_FATAL',
                                     '2 - TRACE_LEVEL_ERROR',
                                     '3 - TRACE_LEVEL_WARNING',
                                     '4 - TRACE_LEVEL_INFORMATION',
                                     '5 - TRACE_LEVEL_VERBOSE');

{ TTraceData }

// Cre et initialise une nouvelle instance de la classe TTraceData.
// <pEvent> est l'vnement de la trace ETW tel qu'il est fournit  la
// mthode callback lors du traitement de ProcessTrace.
constructor TTraceData.Create(pEvent: PEVENT_TRACE);
begin
  inherited Create(pEvent);
  // On prpare 1 octet de plus que le champ MOF rellement fournit.
  SetLength(UserData, pEvent^.MofLength+1);

  // Recopie des infos MOF dans UserData.
  move(pEvent^.MofData^, UserData[0], pEvent^.MofLength);
  UserData[pEvent^.MofLength] := 0; // On ajoute un 0 terminal pour les chanes.
end;

// Retourne l'instance TWmiObject qui dcrit la structure des donnes
// utilisateur dans UserData. Si les informations MOF n'ont pas t publies
// la mthode renvoie nil.
function TTraceData.GetEventClass: string;
begin
  result := FEventName;
end;

function TTraceData.GetEventType: string;
begin
  result := FEventTypeName;
end;

function TTraceData.GetInfo: TWmiObject;
begin
  // Les informations MOF sont dfinies pour une classe d'vnement, un numro
  // de version et un type d'vnement.
  result := WmiInfo.GetEventInfo(EventClass, EventVersion, EventType);
end;

// Retourne les informations utilisateur, formatte sous la forme d'une
// chane de caractres.
function TTraceData.GetUserData: string;
var i : integer;
begin
  // FUserData est prcalcul par la mthode Setup. Si ce n'est pas le cas,
  // ou si les informations MOF taient absentes, on calcul un dump hexadcimal
  // de UserData.
  if FUserData = ''
  then begin
    for i := low(UserData) to high(UserData) do
    begin
      FUserData := FUserData + Format('%0.2x', [UserData[i]]);
    end;
  end;
  result := FUserData;
end;

// Fonction utilitaire qui dcode les donnes de UserData et les formate
// pour qu'elles soient renvoye par les autres mthodes de la classe.
procedure TTraceData.Setup;
var wmiObj : TWmiObject;
begin
  if not FSetup
  then begin
    wmiObj := WmiInfo.GetEventInfo(EventClass, EventVersion, EventType);
    if Assigned(wmiObj)
    then begin
      FEventName := wmiObj.Description;
      FEventTypeName := wmiObj.GetEventTypeName(EventType);
      FUserData := WmiObj.GetUser(@UserData[0], high(UserData)+1);
    end
    else begin
      FEventName := GuidToString(EventClass);
      FEventTypeName := IntToStr(EventType);
      FUserData := FormatHex(@UserData[0], high(UserData)+1);
    end;
    FSetup := true;
  end;
end;

end.
